# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.objstorage.api.client import RemoteObjStorage
from swh.objstorage.backends.pathslicing import PathSlicingObjStorage
from swh.objstorage.factory import get_objstorage


def test_pathslicing_objstorage(tmpdir):
    conf = {"cls": "pathslicing", "root": tmpdir, "slicing": "0:2/0:5"}
    st = get_objstorage(**conf)
    assert isinstance(st, PathSlicingObjStorage)


def test_remote_objstorage():
    conf = {"cls": "remote", "url": "http://127.0.0.1:4242/"}
    st = get_objstorage(**conf)
    assert isinstance(st, RemoteObjStorage)



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import queue
import threading
from typing import Dict, Iterable, Iterator, List, Mapping, Optional, Set, Tuple, Union

# note: it's required to access the statsd object form the statsd module to
# help mocking it in tests...
from swh.core import statsd
from swh.model.model import Sha1
from swh.objstorage.exc import (
    NoBackendsLeftError,
    ObjCorruptedError,
    ObjNotFoundError,
    ObjStorageAPIError,
)
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import CompositeObjId, ObjId, ObjStorageInterface
from swh.objstorage.objstorage import ObjStorage, timed
from swh.objstorage.utils import format_obj_id

MP_COUNTER_METRICS = "swh_objstorage_multiplexer_backend_total"
MP_BACKEND_DISABLED_METRICS = "swh_objstorage_multiplexer_backend_disabled_total"
MP_BACKEND_ENABLED_METRICS = "swh_objstorage_multiplexer_backend_enabled_total"

logger = logging.getLogger(__name__)

DEFAULT_TRANSIENT_READ_EXCEPTIONS = (
    "requests.exceptions.ConnectTimeout",
    "requests.exceptions.ReadTimeout",
    "requests.exceptions.SSLError",
    "requests.exceptions.ConnectionError",
    "builtins.TimeoutError",
    "builtins.IOError",
)


class ObjStorageThread(threading.Thread):
    def __init__(self, storage):
        super().__init__(daemon=True)
        self.storage = storage
        self.commands = queue.Queue()

    def run(self):
        while True:
            try:
                mailbox, command, args, kwargs = self.commands.get(True, 0.05)
            except queue.Empty:
                continue

            try:
                ret = getattr(self.storage, command)(*args, **kwargs)
            except Exception as exc:
                self.queue_result(mailbox, "exception", exc)
            else:
                self.queue_result(mailbox, "result", ret)

    def queue_command(self, command, *args, mailbox=None, **kwargs):
        """Enqueue a new command to be processed by the thread.

        Args:
          command (str): one of the method names for the underlying storage.
          mailbox (queue.Queue): explicit mailbox if the calling thread wants
            to override it.
          args, kwargs: arguments for the command.

        Returns:
          queue.Queue: The mailbox you can read the response from
        """
        if not mailbox:
            mailbox = queue.Queue()
        self.commands.put((mailbox, command, args, kwargs))
        return mailbox

    def queue_result(self, mailbox, result_type, result):
        """Enqueue a new result in the mailbox

        This also provides a reference to the storage, which can be useful when
        an exceptional condition arises.

        Args:
          mailbox (queue.Queue): the mailbox to which we need to enqueue the
            result
          result_type (str): one of 'result', 'exception'
          result: the result to pass back to the calling thread
        """
        mailbox.put(
            {
                "type": result_type,
                "result": result,
            }
        )

    @staticmethod
    def get_result_from_mailbox(mailbox, *args, **kwargs):
        """Unpack the result from the mailbox.

        Args:
          mailbox (queue.Queue): A mailbox to unpack a result from
          args: positional arguments to :func:`mailbox.get`
          kwargs: keyword arguments to :func:`mailbox.get`

        Returns:
          the next result unpacked from the queue

        Raises:
          either the exception we got back from the underlying storage,
            or :exc:`queue.Empty` if :func:`mailbox.get` raises that.
        """

        result = mailbox.get(*args, **kwargs)
        if result["type"] == "exception":
            raise result["result"] from None
        else:
            return result["result"]

    @staticmethod
    def collect_results(mailbox, num_results):
        """Collect num_results from the mailbox"""
        collected = 0
        ret = []
        while collected < num_results:
            try:
                ret.append(
                    ObjStorageThread.get_result_from_mailbox(mailbox, True, 0.05)
                )
            except queue.Empty:
                continue
            collected += 1
        return ret

    def __getattr__(self, attr):
        def call(*args, **kwargs):
            mailbox = self.queue_command(attr, *args, **kwargs)
            return self.get_result_from_mailbox(mailbox)

        return call

    def __contains__(self, *args, **kwargs):
        mailbox = self.queue_command("__contains__", *args, **kwargs)
        return self.get_result_from_mailbox(mailbox)


class MultiplexerObjStorage(ObjStorage):
    """Implementation of ObjStorage that distributes between multiple
    storages.

    The multiplexer object storage allows an input to be demultiplexed
    among multiple storages that will or will not accept it by
    themselves.

    As the ids can be different, no pre-computed ids should be
    submitted.  Also, there are no guarantees that the returned ids
    can be used directly into the storages that the multiplexer
    manage.

    Use case examples follow.

    Example::

        storage_v1 = ReadOnlyProxyObjStorage(
                        PathSlicingObjStorage('/dir1', '0:2/2:4/4:6')
        )
        storage_v2 = PathSlicingObjStorage('/dir2', '0:1/0:5')
        storage = MultiplexerObjStorage([storage_v1, storage_v2])

    When using 'storage', all the new contents will only be added to the v2
    storage, while it will be retrievable from both.

    """

    name: str = "multiplexer"

    def __init__(
        self,
        *,
        objstorages: Iterable[Union[ObjStorageInterface, Dict]],
        read_exception_cooldown: float = 5,
        transient_read_exceptions: Optional[List[str]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.storages = [
            get_objstorage(**sto) if isinstance(sto, dict) else sto
            for sto in objstorages
        ]
        self.storage_threads = [ObjStorageThread(storage) for storage in self.storages]
        for thread in self.storage_threads:
            thread.start()
        self.write_storage_threads = [
            thread
            for thread, storage in zip(self.storage_threads, self.storages)
            if storage.check_config(check_write=True)
        ]

        self.read_exception_cooldown = read_exception_cooldown
        self.transient_read_exceptions = set(
            transient_read_exceptions or DEFAULT_TRANSIENT_READ_EXCEPTIONS
        )

        self.active_readers: Set[int] = set()
        self.reset_timers: Dict[int, threading.Timer] = {}
        self.reset_active_readers()

    def reset_active_readers(self):
        """Reset the active readers set to all storages, and cancel all reset_failed_threads"""
        for t in self.reset_timers.values():
            if t.is_alive():
                t.cancel()

        self.reset_timers = {}
        self.active_readers = set(range(len(self.storages)))

    def disable_backend(self, endpoint: str, name: str, i: int) -> None:
        """Mark read backend `name` at index `i` as failed, from endpoint `endpoint`."""
        if i not in self.active_readers:
            return

        statsd.statsd.increment(
            MP_BACKEND_DISABLED_METRICS,
            1,
            tags={
                "endpoint": endpoint,
                "name": self.name,
                "backend": name,
                "backend_number": i,
            },
        )

        self.active_readers.remove(i)
        if i not in self.reset_timers:
            reset_failed_timer = threading.Timer(
                self.read_exception_cooldown,
                self.enable_backend,
                kwargs={"name": name, "i": i},
            )
            reset_failed_timer.start()
            self.reset_timers[i] = reset_failed_timer

        if not self.active_readers:
            raise NoBackendsLeftError(
                "All backends disabled due to transient failures!"
            )

    def enable_backend(self, name: str, i: int):
        """Mark a reader as available again"""
        statsd.statsd.increment(
            MP_BACKEND_ENABLED_METRICS,
            1,
            tags={
                "name": self.name,
                "backend": name,
                "backend_number": i,
            },
        )

        self.active_readers.add(i)
        if i in self.reset_timers:
            del self.reset_timers[i]

    def wrap_call(self, threads, call, *args, **kwargs):
        threads = list(threads)
        mailbox = queue.Queue()
        for thread in threads:
            thread.queue_command(call, *args, mailbox=mailbox, **kwargs)

        return ObjStorageThread.collect_results(mailbox, len(threads))

    def get_read_threads(self, obj_id=None):
        yield from self.storage_threads

    def get_write_threads(self, obj_id=None):
        yield from self.write_storage_threads

    def check_config(self, *, check_write):
        """Check whether the object storage is properly configured.

        If check_write is True, return True if at least one object storage
        returned True.

        Args:
            check_write (bool): if True, check if writes to the object storage
            can succeed.

        Returns:
            True if the configuration check worked, an exception if it didn't.

        """
        if not check_write:
            return all(
                self.wrap_call(
                    self.storage_threads, "check_config", check_write=check_write
                )
            )
        else:
            return any(
                self.wrap_call(
                    self.storage_threads, "check_config", check_write=check_write
                )
            )

    @timed
    def __contains__(self, obj_id: ObjId) -> bool:
        """Indicate if the given object is present in the storage.

        Args:
            obj_id (bytes): object identifier.

        Returns:
            True if and only if the object is present in the current object
            storage.

        """
        for storage in self.get_read_threads(obj_id):
            if obj_id in storage:
                return True
        return False

    def __iter__(self) -> Iterator[CompositeObjId]:
        def obj_iterator():
            for i, storage in enumerate(self.storages):
                if i in self.active_readers:
                    yield from storage

        return obj_iterator()

    @timed
    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        """Add a new object to the object storage.

        If the adding step works in all the storages that accept this content,
        this is a success. Otherwise, the full adding step is an error even if
        it succeed in some of the storages.

        Args:
            content: content of the object to be added to the storage.
            obj_id: checksum of [bytes] using [ID_HASH_ALGO] algorithm. When
                given, obj_id will be trusted to match the bytes. If missing,
                obj_id will be computed on the fly.
            check_presence: indicate if the presence of the content should be
                verified before adding the file.

        Returns:
            an id of the object into the storage. As the write-storages are
            always readable as well, any id will be valid to retrieve a
            content.
        """
        # note: we do not have per-backend statsd metrics here because the
        # threading scaffolding to manage IO with backends makes it a bit
        # harder to do in a nice manner; plus metrics should be available in
        # the backend objstorages themselves.
        self.wrap_call(
            self.get_write_threads(obj_id),
            "add",
            content,
            obj_id=obj_id,
            check_presence=check_presence,
        )

    def add_batch(
        self,
        contents: Union[Mapping[Sha1, bytes], Iterable[Tuple[ObjId, bytes]]],
        check_presence: bool = True,
    ) -> Dict:
        """Add a batch of new objects to the object storage."""
        write_threads = list(self.get_write_threads())
        results = self.wrap_call(
            write_threads,
            "add_batch",
            contents,
            check_presence=check_presence,
        )

        summed = {"object:add": 0, "object:add:bytes": 0}
        for result in results:
            summed["object:add"] += result["object:add"]
            summed["object:add:bytes"] += result["object:add:bytes"]

        return {
            "object:add": summed["object:add"] // len(results),
            "object:add:bytes": summed["object:add:bytes"] // len(results),
        }

    def restore(self, content: bytes, obj_id: ObjId) -> None:
        return self.wrap_call(
            self.get_write_threads(obj_id),
            "restore",
            content,
            obj_id=obj_id,
        ).pop()

    @timed
    def get(self, obj_id: ObjId) -> bytes:
        corrupted_exc: Optional[ObjCorruptedError] = None
        for i, storage in enumerate(self.get_read_threads(obj_id)):
            if i not in self.active_readers:
                continue
            try:
                obj = storage.get(obj_id)
                statsd.statsd.increment(
                    MP_COUNTER_METRICS,
                    1,
                    tags={
                        "endpoint": "get",
                        "name": self.name,
                        "backend": storage.storage.name,
                        "backend_number": i,
                    },
                )
                return obj
            except ObjNotFoundError:
                continue
            except ObjCorruptedError as exc:
                logger.warning(
                    "Object %s was reported as corrupted by backend '%s': %s",
                    format_obj_id(obj_id),
                    storage.storage.name,
                    str(exc),
                )
                # Hoist exception, mypy doesn't like when we directly use
                # `except ObjCorruptedError as corrupted_exc`
                corrupted_exc = exc
                # Try reading from another storage
                continue
            except Exception as exc:
                if isinstance(exc, ObjStorageAPIError):
                    exc = exc.args[0]
                exc_class = f"{exc.__class__.__module__}.{exc.__class__.__name__}"
                if exc_class in self.transient_read_exceptions:
                    logger.warning(
                        "While reading object %s, received transient read "
                        "exception on backend %s, marking backend as failed",
                        format_obj_id(obj_id),
                        storage.storage.name,
                        exc_info=True,
                    )
                    self.disable_backend(endpoint="get", name=storage.storage.name, i=i)
                    continue
                raise

        if corrupted_exc:
            # The only objects we've found were corrupted, raise that exception
            raise corrupted_exc
        else:
            # No storage contains this content, raise the error
            raise ObjNotFoundError(obj_id)

    def check(self, obj_id: ObjId) -> None:
        nb_present = 0
        nb_corrupted = 0
        exception: Optional[Exception] = None
        for storage in self.get_read_threads(obj_id):
            try:
                storage.check(obj_id)
            except ObjNotFoundError as e:
                exception = e
                continue
            except ObjCorruptedError as e:
                exception = e
                nb_corrupted += 1
            else:
                nb_present += 1

        # Raise exception only if the content could not be found in all the storages
        # or is corrupted in all the storages
        if exception and (nb_present == 0 or nb_corrupted == len(self.storages)):
            raise exception

    def delete(self, obj_id: ObjId):
        super().delete(obj_id)  # Check delete permission
        return all(self.wrap_call(self.get_write_threads(obj_id), "delete", obj_id))



# Copyright (C) 2016-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import base64
import collections
from dataclasses import dataclass
import os
import secrets
import shutil
import subprocess
from urllib.parse import parse_qs, urlparse

from azure.core.exceptions import ResourceExistsError, ResourceNotFoundError
from azure.storage.blob import BlobServiceClient
import pytest

from swh.model.hashutil import hash_to_hex
import swh.objstorage.backends.azure
from swh.objstorage.exc import ObjCorruptedError
from swh.objstorage.factory import get_objstorage
from swh.objstorage.objstorage import decompressors

from .objstorage_testing import ObjStorageTestFixture

AZURITE_EXE = shutil.which(
    "azurite-blob", path=os.environ.get("AZURITE_PATH", os.environ.get("PATH"))
)


@dataclass
class MockListedObject:
    name: str


class MockAsyncDownloadClient:
    def __init__(self, blob_data):
        self.blob_data = blob_data

    def content_as_bytes(self):
        future = asyncio.Future()
        future.set_result(self.blob_data)
        return future


class MockDownloadClient:
    def __init__(self, blob_data):
        self.blob_data = blob_data

    def content_as_bytes(self):
        return self.blob_data

    def __await__(self):
        yield from ()
        return MockAsyncDownloadClient(self.blob_data)


class MockBlobClient:
    def __init__(self, container, blob):
        self.container = container
        self.blob = blob

    def get_blob_properties(self):
        if self.blob not in self.container.blobs:
            raise ResourceNotFoundError("Blob not found")

        return {"exists": True}

    def upload_blob(self, data, length=None):
        if self.blob in self.container.blobs:
            raise ResourceExistsError("Blob already exists")

        if length is not None and length != len(data):
            raise ValueError("Wrong length for blob data!")

        self.container.blobs[self.blob] = data

    def download_blob(self):
        if self.blob not in self.container.blobs:
            raise ResourceNotFoundError("Blob not found")

        return MockDownloadClient(self.container.blobs[self.blob])

    def delete_blob(self):
        if self.blob not in self.container.blobs:
            raise ResourceNotFoundError("Blob not found")

        del self.container.blobs[self.blob]


@pytest.fixture(scope="class")
def azurite_connection_string(tmpdir_factory):
    host = "127.0.0.1"

    azurite_path = tmpdir_factory.mktemp("azurite")

    azurite_proc = subprocess.Popen(
        [
            AZURITE_EXE,
            "--blobHost",
            host,
            "--blobPort",
            "0",
        ],
        stdout=subprocess.PIPE,
        cwd=azurite_path,
    )

    prefix = b"Azurite Blob service successfully listens on "
    for line in azurite_proc.stdout:
        if line.startswith(prefix):
            base_url = line[len(prefix) :].decode().strip()
            break
    else:
        assert False, "Did not get Azurite Blob service port."

    # https://learn.microsoft.com/en-us/azure/storage/common/storage-use-azurite#well-known-storage-account-and-key
    account_name = "devstoreaccount1"
    account_key = (
        "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq"
        "/K1SZFPTOtr/KBHBeksoGMGw=="
    )

    container_url = f"{base_url}/{account_name}"
    # note the stripping of the scheme for the BlobSecondaryEndpoint is NOT
    # a mistake; looks like azurite requires this secondary endpoint to not
    # come with the scheme (although I've not seen this documented, so not
    # sure if it's by design or a bug).
    secondary_url = container_url.replace("http://127.0.0.1", "localhost")

    yield (
        f"DefaultEndpointsProtocol=https;"
        f"AccountName={account_name};"
        f"AccountKey={account_key};"
        f"BlobEndpoint={container_url};"
        f"BlobSecondaryEndpoint={secondary_url};"
    )

    azurite_proc.kill()
    azurite_proc.wait(2)


@pytest.mark.skipif(not AZURITE_EXE, reason="azurite not found in AZURITE_PATH or PATH")
class TestAzuriteCloudObjStorage(ObjStorageTestFixture):
    compression = "none"

    @pytest.fixture
    def swh_objstorage_config(self, azurite_connection_string):
        self._container_name = secrets.token_hex(10)
        client = BlobServiceClient.from_connection_string(azurite_connection_string)
        client.create_container(self._container_name)

        return {
            "cls": "azure",
            "connection_string": azurite_connection_string,
            "container_name": self._container_name,
            "compression": self.compression,
        }

    def test_download_url(self, azurite_connection_string):
        content_p, obj_id_p = self.hash_content(b"contains_present")
        self.storage.add(content_p, obj_id=obj_id_p)
        assert self.storage.download_url(obj_id_p).startswith("http://127.0.0.1:")
        storage2 = get_objstorage(
            "azure",
            connection_string=azurite_connection_string,
            container_name=self._container_name,
            compression=self.compression,
            use_secondary_endpoint_for_downloads=True,
        )
        assert storage2.download_url(obj_id_p).startswith("http://localhost:")


class TestAzuriteCloudObjStorageGzip(TestAzuriteCloudObjStorage):
    compression = "gzip"


def get_MockContainerClient():
    blobs = collections.defaultdict(dict)  # {container_url: {blob_id: blob}}

    class MockContainerClient:
        def __init__(self, container_url):
            self.container_url = container_url
            self.blobs = blobs[self.container_url]

        @classmethod
        def from_container_url(cls, container_url):
            return cls(container_url)

        def get_container_properties(self):
            return {"exists": True}

        def get_blob_client(self, blob):
            return MockBlobClient(self, blob)

        def list_blobs(self):
            for obj in sorted(self.blobs):
                yield MockListedObject(obj)

        def delete_blob(self, blob):
            self.get_blob_client(blob.name).delete_blob()

        def __aenter__(self):
            return self

        def __await__(self):
            future = asyncio.Future()
            future.set_result(self)
            yield from future

        def __aexit__(self, *args):
            return self

    return MockContainerClient


class TestMockedAzureCloudObjStorage(ObjStorageTestFixture):
    compression = "none"

    @pytest.fixture
    def swh_objstorage_config(self, mocker):
        ContainerClient = get_MockContainerClient()
        mocker.patch("swh.objstorage.backends.azure.ContainerClient", ContainerClient)

        mocker.patch(
            "swh.objstorage.backends.azure.AsyncContainerClient", ContainerClient
        )

        return {
            "cls": "azure",
            "container_url": "https://bogus-container-url.example",
            "compression": self.compression,
        }

    def test_compression(self):
        content, obj_id = self.hash_content(b"test content is compressed")
        self.storage.add(content, obj_id=obj_id)

        internal_id = self.storage._internal_id(obj_id)
        blob_client = self.storage.get_blob_client(internal_id)
        raw_blob = blob_client.download_blob().content_as_bytes()

        d = decompressors[self.compression]()
        assert d.decompress(raw_blob) == content
        assert d.unused_data == b""

    def test_trailing_data_on_stored_blob(self):
        content, obj_id = self.hash_content(b"test content without garbage")
        self.storage.add(content, obj_id=obj_id)

        internal_id = self.storage._internal_id(obj_id)
        blob_client = self.storage.get_blob_client(internal_id)
        raw_blob = blob_client.download_blob().content_as_bytes()
        new_data = raw_blob + b"trailing garbage"

        blob_client.delete_blob()
        blob_client.upload_blob(data=new_data, length=len(new_data))

        with pytest.raises(ObjCorruptedError) as e:
            self.storage.check(obj_id)

        if self.compression != "none":
            assert "trailing data found" in e.value.args[0]

    @pytest.mark.skip("makes no sense to test this for the mocked azure")
    def test_download_url(self):
        pass


class TestMockedAzureCloudObjStorageGzip(TestMockedAzureCloudObjStorage):
    compression = "gzip"


@pytest.mark.all_compression_methods
class TestMockedAzureCloudObjStorageZlib(TestMockedAzureCloudObjStorage):
    compression = "zlib"


@pytest.mark.all_compression_methods
class TestMockedAzureCloudObjStorageLzma(TestMockedAzureCloudObjStorage):
    compression = "lzma"


@pytest.mark.all_compression_methods
class TestMockedAzureCloudObjStorageBz2(TestMockedAzureCloudObjStorage):
    compression = "bz2"


class TestPrefixedAzureCloudObjStorage(ObjStorageTestFixture):
    @pytest.fixture
    def swh_objstorage_config(self, mocker):
        self.ContainerClient = get_MockContainerClient()
        mocker.patch(
            "swh.objstorage.backends.azure.ContainerClient", self.ContainerClient
        )

        mocker.patch(
            "swh.objstorage.backends.azure.AsyncContainerClient", self.ContainerClient
        )

        self.accounts = {}
        for prefix in "0123456789abcdef":
            self.accounts[prefix] = "https://bogus-container-url.example/" + prefix

        return {"cls": "azure-prefixed", "accounts": self.accounts}

    def test_prefixedazure_instantiation_missing_prefixes(self):
        del self.accounts["d"]
        del self.accounts["e"]

        with pytest.raises(ValueError, match="Missing prefixes"):
            get_objstorage("azure-prefixed", accounts=self.accounts)

    def test_prefixedazure_instantiation_inconsistent_prefixes(self):
        self.accounts["00"] = self.accounts["0"]

        with pytest.raises(ValueError, match="Inconsistent prefixes"):
            get_objstorage("azure-prefixed", accounts=self.accounts)

    def test_prefixedazure_sharding_behavior(self):
        for i in range(100):
            content, obj_id = self.hash_content(b"test_content_%02d" % i)
            self.storage.add(content, obj_id=obj_id)
            hex_obj_id = hash_to_hex(obj_id)
            prefix = hex_obj_id[0]
            assert (
                self.ContainerClient(self.storage.container_urls[prefix])
                .get_blob_client(hex_obj_id)
                .get_blob_properties()
            )

    @pytest.mark.skip("makes no sense to test this for the mocked azure")
    def test_download_url(self):
        pass


def test_get_container_url():
    # r=read, l=list, w=write, d=delete
    policy_map = {
        "read_only": "rl",
        "append_only": "rwl",
        "full": "rwdl",
    }

    for policy, expected in policy_map.items():
        ret = swh.objstorage.backends.azure.get_container_url(
            account_name="account_name",
            account_key=base64.b64encode(b"account_key"),
            container_name="container_name",
            access_policy=policy,
        )

        p = urlparse(ret)
        assert p.scheme == "https"
        assert p.netloc == "account_name.blob.core.windows.net"
        assert p.path == "/container_name"

        qs = parse_qs(p.query)
        # sp: permissions
        assert qs["sp"] == [expected]
        # sr: resource (c=container)
        assert qs["sr"] == ["c"]
        # st: start; se: expiry
        assert qs["st"][0] < qs["se"][0]


def test_bwcompat_args(monkeypatch):
    monkeypatch.setattr(
        swh.objstorage.backends.azure,
        "ContainerClient",
        get_MockContainerClient(),
    )

    with pytest.deprecated_call():
        objs = get_objstorage(
            "azure",
            account_name="account_name",
            api_secret_key=base64.b64encode(b"account_key"),
            container_name="container_name",
        )

    assert objs is not None


def test_bwcompat_args_prefixed(monkeypatch):
    monkeypatch.setattr(
        swh.objstorage.backends.azure,
        "ContainerClient",
        get_MockContainerClient(),
    )

    accounts = {
        prefix: {
            "account_name": f"account_name{prefix}",
            "api_secret_key": base64.b64encode(b"account_key"),
            "container_name": "container_name",
        }
        for prefix in "0123456789abcdef"
    }

    with pytest.deprecated_call():
        objs = get_objstorage("azure-prefixed", accounts=accounts)

    assert objs is not None



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import sys

from matplotlib import pyplot as plt
from matplotlib.ticker import FormatStrFormatter
import pandas as pd


def human(size, unit):
    if size < 1024:
        return f"{int(size)} {unit}/s"
    elif size / 1024 < 1024:
        return f"{round(size/1024, 1)} K{unit}/s"
    elif size / (1024 * 1024) < 1024:
        return f"{round(size / (1024 * 1024), 1)} M{unit}/s"
    elif size / (1024 * 1024 * 1024) < 1024:
        return f"{round(size / (1024 * 1024 * 1024), 1)} G{unit}/s"


def read_stats(stats):
    dfs = []
    files = os.listdir(stats)
    for file in files:
        f = f"{stats}/{file}"
        if not os.path.isfile(f):
            continue
        dfs.append(pd.read_csv(f))
    df = pd.concat(dfs)
    df.set_index("time")
    return df.sort_values(by=["time"])


def main(stats):
    df = read_stats(stats)
    print(df)
    t = df["time"].to_numpy()
    sec = t[-1] - t[0]
    a = df.sum() / sec
    print(f'Bytes write   {human(a["bytes_write"], "B")}')
    print(f'Objects write {human(a["object_write_count"], "object")}')
    print(f'Bytes read    {human(a["bytes_read"], "B")}')
    print(f'Objects read  {human(a["object_read_count"], "object")}')

    df["date"] = pd.to_datetime(df["time"], unit="s")

    p = df.plot(x="time", y=["bytes_write", "bytes_read"])
    p.set_xlabel("Time")
    p.yaxis.set_major_formatter(FormatStrFormatter("%.0f"))
    p.set_ylabel("B/s")
    plt.show()


if __name__ == "__main__":
    main(sys.argv[1])



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from itertools import permutations

import pytest

from swh.objstorage.interface import (
    COMPOSITE_OBJID_KEYS,
    CompositeObjId,
    objid_from_dict,
)


def test_composite_objid_keys():
    assert (
        CompositeObjId.__required_keys__ | CompositeObjId.__optional_keys__
        == COMPOSITE_OBJID_KEYS
    )


def test_objid_from_dict_missing_key():
    with pytest.raises(ValueError, match="missing at least one of") as excinfo:
        _ = objid_from_dict({"foo": "bar"})

    for key in COMPOSITE_OBJID_KEYS:
        assert key in excinfo.value.args[0]


def test_objid_from_dict_valueerror():
    for key in COMPOSITE_OBJID_KEYS:
        with pytest.raises(TypeError, match=f"value for {key} is str, not bytes"):
            _ = objid_from_dict({key: "boom"})


def test_objid_from_dict():
    for i in range(len(COMPOSITE_OBJID_KEYS)):
        for j, keys in enumerate(permutations(COMPOSITE_OBJID_KEYS, i + 1)):
            d = {
                key: int.to_bytes(i * 1500 + j * 50 + k, 20, "little")
                for k, key in enumerate(keys)
            }

            objid = objid_from_dict(d)

            assert objid == d

            d2 = {**d, f"test{i:03x}": "garbage", f"extra{j:02d}": "extra"}

            assert objid_from_dict(d2) == d



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.core.pytest_plugin import FakeSocket
from swh.objstorage.factory import get_objstorage


@pytest.fixture
def swh_objstorage_config():
    """Fixture that returns a dictionary containing the configuration
    required to instantiate an object storage.

    Unless the fixture gets overridden, the configuration for creating an
    object storage in memory is returned.

    See :func:`swh.objstorage.factory.get_objstorage` for more details.
    """
    return {"cls": "memory"}


@pytest.fixture
def swh_objstorage(swh_objstorage_config):
    """Fixture that instantiates an object storage based on the configuration
    returned by the ``swh_objstorage_config`` fixture.
    """
    return get_objstorage(**swh_objstorage_config)


@pytest.fixture
def statsd():
    """Simple fixture giving a Statsd instance suitable for tests

    It will replace the `swh.core.statsd.statsd` instance with this one.

    The Statsd instance uses a FakeSocket as `.socket` attribute in which one
    can get the accumulated statsd messages in a deque in `.socket.payloads`.
    """

    import swh.core.statsd

    statsd = swh.core.statsd.Statsd()
    statsd._socket = FakeSocket()
    swh.core.statsd.statsd = statsd
    yield statsd



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
from itertools import islice
import logging
import os
import tempfile
from typing import Iterator, List, Literal, Optional

from swh.objstorage.constants import (
    DEFAULT_LIMIT,
    ID_HASH_ALGO,
    ID_HEXDIGEST_LENGTH,
    is_valid_hexdigest,
)
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.interface import CompositeObjId, ObjId
from swh.objstorage.objstorage import (
    CompressionFormat,
    ObjStorage,
    compressors,
    objid_to_default_hex,
    timed,
)

BUFSIZ = 1048576

DIR_MODE = 0o755
FILE_MODE = 0o644


logger = logging.getLogger(__name__)


def is_valid_filename(filename: str, algo: Literal["sha1", "sha256"] = ID_HASH_ALGO):
    """Checks that the file points to a valid hexdigest for the given algo."""

    return is_valid_hexdigest(os.path.basename(filename), algo)


class PathSlicer:
    """Helper class to compute a path based on a hash.

    Used to compute a directory path based on the object hash according to a
    given slicing. Each slicing correspond to a directory that is named
    according to the hash of its content.

    For instance a file with SHA1 34973274ccef6ab4dfaaf86599792fa9c3fe4689
    will have the following computed path:

    - 0:2/2:4/4:6 : 34/97/32/34973274ccef6ab4dfaaf86599792fa9c3fe4689
    - 0:1/0:5/    : 3/34973/34973274ccef6ab4dfaaf86599792fa9c3fe4689

     Args:
         root (str): path to the root directory of the storage on the disk.
         slicing (str): the slicing configuration.
    """

    def __init__(self, root: str, slicing: str):
        self.root = root
        # Make a list of tuples where each tuple contains the beginning
        # and the end of each slicing.
        try:
            self.bounds = [
                slice(*(int(x) if x else None for x in sbounds.split(":")))
                for sbounds in slicing.split("/")
                if sbounds
            ]
        except TypeError:
            raise ValueError(
                "Invalid slicing declaration; "
                "it should be a of the form '<int>:<int>[/<int>:<int>]..."
            )

    def check_config(self):
        """Check the slicing configuration is valid.

        Raises:
            ValueError: if the slicing configuration is invalid.
        """
        if len(self):
            max_char = max(
                max(bound.start or 0, bound.stop or 0) for bound in self.bounds
            )
            if ID_HEXDIGEST_LENGTH < max_char:
                raise ValueError(
                    "Algorithm %s has too short hash for slicing to char %d"
                    % (ID_HASH_ALGO, max_char)
                )

    def get_directory(self, hex_obj_id: str) -> str:
        """Compute the storage directory of an object.

        See also: PathSlicer::get_path

        Args:
            hex_obj_id: object id as hexlified string.

        Returns:
            Absolute path (including root) to the directory that contains
            the given object id.
        """
        return os.path.join(self.root, *self.get_slices(hex_obj_id))

    def get_path(self, hex_obj_id: str) -> str:
        """Compute the full path to an object into the current storage.

        See also: PathSlicer::get_directory

        Args:
            hex_obj_id(str): object id as hexlified string.

        Returns:
            Absolute path (including root) to the object corresponding
            to the given object id.
        """
        return os.path.join(self.get_directory(hex_obj_id), hex_obj_id)

    def get_slices(self, hex_obj_id: str) -> List[str]:
        """Compute the path elements for the given hash.

        Args:
            hex_obj_id(str): object id as hexlified string.

        Returns:
            Relative path to the actual object corresponding to the given id as
            a list.
        """

        assert len(hex_obj_id) == ID_HEXDIGEST_LENGTH
        return [hex_obj_id[bound] for bound in self.bounds]

    def __len__(self) -> int:
        """Number of slices of the slicer"""
        return len(self.bounds)


class PathSlicingObjStorage(ObjStorage):
    """Implementation of the ObjStorage API based on the hash of the content.

    On disk, an object storage is a directory tree containing files
    named after their object IDs. An object ID is a checksum of its
    content, depending on the value of the ID_HASH_ALGO constant (see
    swh.model.hashutil for its meaning).

    To avoid directories that contain too many files, the object storage has a
    given slicing. Each slicing correspond to a directory that is named
    according to the hash of its content.

    So for instance a file with SHA1 34973274ccef6ab4dfaaf86599792fa9c3fe4689
    will be stored in the given object storages :

    - 0:2/2:4/4:6 : 34/97/32/34973274ccef6ab4dfaaf86599792fa9c3fe4689
    - 0:1/0:5/    : 3/34973/34973274ccef6ab4dfaaf86599792fa9c3fe4689

    The files in the storage are stored in gzipped compressed format.

    Args:
        root (str): path to the root directory of the storage on
            the disk.
        slicing (str): string that indicates the slicing to perform
            on the hash of the content to know the path where it should
            be stored (see the documentation of the PathSlicer class).

    """

    PRIMARY_HASH: Literal["sha1"] = "sha1"
    name: str = "pathslicing"

    def __init__(
        self,
        *,
        root: str = "",
        compression: CompressionFormat = "gzip",
        slicing: str = "",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.root = root
        self.slicer = PathSlicer(root, slicing)

        self.use_fdatasync = hasattr(os, "fdatasync")
        self.compression = compression

        self.check_config(check_write=False)

    def check_config(self, *, check_write):
        """Check whether this object storage is properly configured"""

        self.slicer.check_config()

        if not os.path.isdir(self.root):
            raise ValueError(
                'PathSlicingObjStorage root "%s" is not a directory' % self.root
            )

        if check_write:
            if not os.access(self.root, os.W_OK):
                raise PermissionError(
                    'PathSlicingObjStorage root "%s" is not writable' % self.root
                )

        if self.compression not in compressors:
            raise ValueError(
                'Unknown compression algorithm "%s" for '
                "PathSlicingObjStorage" % self.compression
            )

        return True

    @timed
    def __contains__(self, obj_id: ObjId) -> bool:
        hex_obj_id = objid_to_default_hex(obj_id)
        return os.path.isfile(self.slicer.get_path(hex_obj_id))

    def __iter__(self) -> Iterator[CompositeObjId]:
        """Iterate over the object identifiers currently available in the
        storage.

        Warning: with the current implementation of the object
        storage, this method will walk the filesystem to list objects,
        meaning that listing all objects will be very slow for large
        storages. You almost certainly don't want to use this method
        in production.

        Return:
            Iterator over object IDs

        """

        # XXX hackish: it does not verify that the depth of found files
        # matches the slicing depth of the storage
        for root, _dirs, files in os.walk(self.root):
            _dirs.sort()
            for f in sorted(files):
                if not is_valid_filename(f, algo=self.PRIMARY_HASH):
                    logger.warning("__iter__ skipping over invalid file %s/%s", root, f)
                    continue
                yield {self.PRIMARY_HASH: bytes.fromhex(f)}

    def __len__(self) -> int:
        """Compute the number of objects available in the storage.

        Warning: this currently uses `__iter__`, its warning about bad
        performances applies

        Return:
            number of objects contained in the storage
        """
        return sum(1 for i in self)

    @timed
    def add(
        self,
        content: bytes,
        obj_id: ObjId,
        check_presence: bool = True,
    ) -> None:
        if check_presence and obj_id in self:
            # If the object is already present, return immediately.
            return

        hex_obj_id = objid_to_default_hex(obj_id)
        with self._write_obj_file(hex_obj_id) as f:
            f.write(self.compress(content))

    @timed
    def get(self, obj_id: ObjId) -> bytes:
        if obj_id not in self:
            raise ObjNotFoundError(obj_id)

        # Open the file and return its content as bytes
        hex_obj_id = objid_to_default_hex(obj_id)
        with open(self.slicer.get_path(hex_obj_id), "rb") as f:
            return self.decompress(f.read(), hex_obj_id)

    def delete(self, obj_id: ObjId):
        super().delete(obj_id)  # Check delete permission
        if obj_id not in self:
            raise ObjNotFoundError(obj_id)

        hex_obj_id = objid_to_default_hex(obj_id)
        try:
            os.remove(self.slicer.get_path(hex_obj_id))
        except FileNotFoundError:
            raise ObjNotFoundError(obj_id)
        return True

    # Streaming methods

    def list_content(
        self, last_obj_id: Optional[ObjId] = None, limit: Optional[int] = DEFAULT_LIMIT
    ) -> Iterator[CompositeObjId]:
        if last_obj_id:
            it = self.iter_from(last_obj_id)
        else:
            it = iter(self)
        return islice(it, limit)

    def iter_from(self, obj_id, n_leaf=False):
        hex_obj_id = objid_to_default_hex(obj_id)
        slices = self.slicer.get_slices(hex_obj_id)
        rlen = len(self.root.split("/"))

        i = 0
        for root, dirs, files in os.walk(self.root):
            if not dirs:
                i += 1
            level = len(root.split("/")) - rlen
            dirs.sort()
            if dirs and root == os.path.join(self.root, *slices[:level]):
                cslice = slices[level]
                for d in dirs[:]:
                    if d < cslice:
                        dirs.remove(d)
            for f in sorted(files):
                if not is_valid_filename(f, algo=self.PRIMARY_HASH):
                    logger.warning(
                        "iter_from skipping over invalid file %s/%s", root, f
                    )
                    continue
                if f > hex_obj_id:
                    yield {self.PRIMARY_HASH: bytes.fromhex(f)}
        if n_leaf:
            yield i

    @contextmanager
    def _write_obj_file(self, hex_obj_id):
        """Context manager for writing object files to the object storage.

        During writing, data are written to a temporary file, which is atomically
        renamed to the right file name after closing.

        Usage sample:
            with objstorage._write_obj_file(hex_obj_id):
                f.write(obj_data)

        Yields:
            a file-like object open for writing bytes.
        """
        # Get the final paths and create the directory if absent.
        dir = self.slicer.get_directory(hex_obj_id)
        if not os.path.isdir(dir):
            os.makedirs(dir, DIR_MODE, exist_ok=True)
        path = os.path.join(dir, hex_obj_id)

        # Create a temporary file.
        (tmp, tmp_path) = tempfile.mkstemp(suffix=".tmp", prefix="hex_obj_id.", dir=dir)

        # Open the file and yield it for writing.
        tmp_f = os.fdopen(tmp, "wb")
        yield tmp_f

        # Make sure the contents of the temporary file are written to disk
        tmp_f.flush()
        if self.use_fdatasync:
            os.fdatasync(tmp)
        else:
            os.fsync(tmp)

        # Then close the temporary file and move it to the right path.
        tmp_f.close()
        os.chmod(tmp_path, FILE_MODE)
        os.rename(tmp_path, path)



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import requests
import requests_mock
from requests_mock.contrib import fixture

from swh.model import hashutil
from swh.objstorage.exc import (
    NonIterableObjStorageError,
    ObjCorruptedError,
    ObjNotFoundError,
    ReadOnlyObjStorageError,
)
from swh.objstorage.factory import get_objstorage
from swh.objstorage.objstorage import compute_hash


def build_objstorage(multi_hash=False):
    """Build an HTTPReadOnlyObjStorage suitable for tests

    this instancaite 2 ObjStorage, one HTTPReadOnlyObjStorage (the "front" one
    being under test), and one InMemoryObjStorage (which actually stores the
    test content), and install a request mock fixture to route HTTP requests
    from the HTTPReadOnlyObjStorage to query the InMemoryStorage.

    Also fills the backend storage with a 100 objects.
    """
    sto_back = get_objstorage(cls="memory")
    objids = []
    for i in range(100):
        content = f"some content {i}".encode()
        if multi_hash:
            obj_id = hashutil.MultiHash.from_data(content).digest()
        else:
            obj_id = compute_hash(content)
        objids.append(obj_id)
        sto_back.add(content, obj_id=obj_id)

    url = "http://127.0.0.1/content/"
    sto_front = get_objstorage(cls="http", url=url)
    mock = fixture.Fixture()
    mock.setUp()

    def get_cb(request, context):
        dirname, basename = request.path.rsplit("/", 1)
        objid = bytes.fromhex(basename)
        if dirname == "/content" and objid in sto_back:
            return sto_back.get(objid)
        context.status_code = 404

    def head_cb(request, context):
        dirname, basename = request.path.rsplit("/", 1)
        objid = bytes.fromhex(basename)
        if dirname != "/content" or objid not in sto_back:
            context.status_code = 404
            return b"Not Found"
        return b"Found"

    mock.register_uri(requests_mock.GET, requests_mock.ANY, content=get_cb)
    mock.register_uri(requests_mock.HEAD, requests_mock.ANY, content=head_cb)

    return sto_front, sto_back, objids


@pytest.mark.parametrize("multi_hash", [False, True])
def test_http_objstorage(multi_hash):
    sto_front, sto_back, objids = build_objstorage(multi_hash)

    for objid in objids:
        assert objid in sto_front
        assert sto_front.get(objid) == sto_back.get(objid)
        assert sto_front.get(objid).decode().startswith("some content ")


def test_http_objstorage_missing():
    sto_front, sto_back, objids = build_objstorage()

    assert b"\x00" * 20 not in sto_front


def test_http_objstorage_get_missing():
    sto_front, sto_back, objids = build_objstorage()

    with pytest.raises(ObjNotFoundError):
        sto_front.get(b"\x00" * 20)


@pytest.mark.parametrize("multi_hash", [False, True])
def test_http_objstorage_check(multi_hash):
    sto_front, sto_back, objids = build_objstorage(multi_hash)
    for objid in objids:
        assert sto_front.check(objid) is None  # no Exception means OK

    # create an invalid object in the in-memory objstorage
    invalid_content = b"p0wn3d content"
    fake_objid = b"\x01" * 20
    sto_back.add(invalid_content, fake_objid)

    # the http objstorage should report it as invalid
    with pytest.raises(ObjCorruptedError):
        sto_front.check(fake_objid)


def test_http_objstorage_read_only():
    sto_front, sto_back, objids = build_objstorage()

    content = b""
    obj_id = compute_hash(content)
    with pytest.raises(ReadOnlyObjStorageError):
        sto_front.add(content, obj_id=obj_id)
    with pytest.raises(ReadOnlyObjStorageError):
        sto_front.restore(b"", obj_id=compute_hash(b""))
    with pytest.raises(ReadOnlyObjStorageError):
        sto_front.delete(b"\x00" * 20)


def test_http_objstorage_not_iterable():
    sto_front, sto_back, objids = build_objstorage()

    with pytest.raises(NonIterableObjStorageError):
        len(sto_front)
    with pytest.raises(NonIterableObjStorageError):
        iter(sto_front)


def test_http_cannonical_url():
    url = "http://127.0.0.1/content"
    sto = get_objstorage(cls="http", url=url)
    assert sto.root_path == url + "/"


@pytest.mark.parametrize("multi_hash", [False, True])
def test_http_objstorage_download_url(multi_hash):
    sto_front, sto_back, objids = build_objstorage(multi_hash)

    for objid in objids:
        assert objid in sto_front
        response = requests.get(sto_front.download_url(objid))
        assert response.text.startswith("some content ")



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from datetime import timedelta
from typing import (
    Any,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from typing_extensions import Literal, Protocol, TypedDict, runtime_checkable

from swh.core.api import remote_api_endpoint
from swh.model.model import Sha1
from swh.objstorage.constants import DEFAULT_LIMIT

COMPOSITE_OBJID_KEYS: FrozenSet[
    Literal["sha1", "sha1_git", "sha256", "blake2s256"]
] = frozenset(("sha1", "sha1_git", "sha256", "blake2s256"))


class CompositeObjId(TypedDict, total=False):
    sha1: bytes
    sha1_git: bytes
    sha256: bytes
    blake2s256: bytes


def objid_from_dict(d: Dict[str, Any]) -> CompositeObjId:
    filtered: CompositeObjId = {}

    for key in COMPOSITE_OBJID_KEYS:
        value = d.get(key)
        if value is None:
            continue
        if not isinstance(value, bytes):
            raise TypeError(
                "value for %s is %s, not bytes" % (key, value.__class__.__name__)
            )
        filtered[key] = value

    if not filtered:
        raise ValueError(
            "dict is missing at least one of %s" % ", ".join(COMPOSITE_OBJID_KEYS)
        )

    return filtered


ObjId = Union[bytes, CompositeObjId]
"""Type of object ids, which should be ``{hash: value for hash in SUPPORTED_HASHES}``;
but single sha1 hashes are supported for legacy clients"""


@runtime_checkable
class ObjStorageInterface(Protocol):
    """High-level API to manipulate the Software Heritage object storage.

    Conceptually, the object storage offers the following methods:

    - check_config()  check if the object storage is properly configured
    - __contains__()  check if an object is present, by object id
    - add()           add a new object, returning an object id
    - restore()       same as add() but erase an already existed content
    - get()           retrieve the content of an object, by object id
    - check()         check the integrity of an object, by object id
    - delete()        remove an object

    and the following attributes:

    - name            name given to the object storage; useful e.g. for logging in
                      composite object storagges (multiplexer)

    Each implementation of this interface can have a different behavior and
    its own way to store the contents.
    """

    name: str

    def __init__(
        self,
        *,
        name: str = "",
        **kwargs,
    ):
        ...

    @remote_api_endpoint("check_config")
    def check_config(self, *, check_write):
        """Check whether the object storage is properly configured.

        Args:
            check_write (bool): if True, check if writes to the object storage
            can succeed.

        Returns:
            True if the configuration check worked, False if 'check_write' is
            True and the object storage is actually read only, and an exception
            if the check failed.

        """
        ...

    @remote_api_endpoint("content/contains")
    def __contains__(self, obj_id: ObjId) -> bool:
        """Indicate if the given object is present in the storage.

        Args:
            obj_id: object identifier.

        Returns:
            True if and only if the object is present in the current object
            storage.

        """
        ...

    @remote_api_endpoint("content/add")
    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        """Add a new object to the object storage.

        Args:
            content: object's raw content to add in storage.
            obj_id: either dict of checksums, or single checksum of
                [bytes] using [ID_HASH_ALGO] algorithm.
                It is trusted to match the bytes.
            check_presence (bool): indicate if the presence of the
                content should be verified before adding the file.

        Returns:
            the id (bytes) of the object into the storage.

        """
        ...

    @remote_api_endpoint("content/add/batch")
    def add_batch(
        self,
        contents: Union[Mapping[Sha1, bytes], Iterable[Tuple[ObjId, bytes]]],
        check_presence: bool = True,
    ) -> Dict:
        """Add a batch of new objects to the object storage.

        Args:
            contents: either mapping from [ID_HASH_ALGO] checksums to object contents,
                or list of pairs of dict hashes and object contents

        Returns:
            the summary of objects added to the storage (count of object,
            count of bytes object)

        """
        ...

    def restore(self, content: bytes, obj_id: ObjId) -> None:
        """Restore a content that have been corrupted.

        This function is identical to add but does not check if
        the object id is already in the file system.
        The default implementation provided by the current class is
        suitable for most cases.

        Args:
            content: object's raw content to add in storage
            obj_id: dict of hashes of the content (or only the sha1, for legacy clients)
        """
        ...

    @remote_api_endpoint("content/get")
    def get(self, obj_id: ObjId) -> bytes:
        """Retrieve the content of a given object.

        Args:
            obj_id: object id.

        Returns:
            the content of the requested object as bytes.

        Raises:
            ObjNotFoundError: if the requested object is missing.

        """
        ...

    @remote_api_endpoint("content/get/batch")
    def get_batch(self, obj_ids: Iterable[ObjId]) -> Iterator[Optional[bytes]]:
        """Retrieve objects' raw content in bulk from storage.

        Note: This function does have a default implementation in
        ObjStorage that is suitable for most cases.

        For object storages that needs to do the minimal number of
        requests possible (ex: remote object storages), that method
        can be overridden to perform a more efficient operation.

        Args:
            obj_ids: list of object ids.

        Returns:
            list of resulting contents, or None if the content could
            not be retrieved. Do not raise any exception as a fail for
            one content will not cancel the whole request.

        """
        ...

    @remote_api_endpoint("content/download_url")
    def download_url(
        self,
        obj_id: ObjId,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        """Get a direct download link for the object if the obstorage backend supports
        such feature.

        Some objstorage backends, typically cloud based ones like azure or s3, can provide
        a direct download link for a stored object.

        Args:
            obj_id: object identifier
            content_disposition: set Content-Disposition header for the generated URL
                response if the objstorage backend supports it
            expiry: the duration after which the URL expires if the objstorage backend
                supports it, if not provided the URL expires 24 hours after its creation

        Returns:
            Direct download URL for the object or :const:`None` if the objstorage backend does
                not support such feature.
        """

        ...

    @remote_api_endpoint("content/check")
    def check(self, obj_id: ObjId) -> None:
        """Perform an integrity check for a given object.

        Verify that the file object is in place and that the content matches
        the object id.

        Args:
            obj_id: object identifier.

        Raises:
            ObjNotFoundError: if the requested object is missing.
            ObjCorruptedError: if the requested object is corrupted.

        """
        ...

    @remote_api_endpoint("content/delete")
    def delete(self, obj_id: ObjId):
        """Delete an object.

        Args:
            obj_id: object identifier.

        Raises:
            ObjNotFoundError: if the requested object is missing.

        """
        ...

    def __iter__(self) -> Iterator[CompositeObjId]:
        ...

    def list_content(
        self, last_obj_id: Optional[ObjId] = None, limit: Optional[int] = DEFAULT_LIMIT
    ) -> Iterator[CompositeObjId]:
        """Generates known object ids.

        Args:
            last_obj_id: object id from which to iterate from
                 (excluded).
            limit (int): max number of object ids to generate. If unset (None),
                 generate all objects (behavior might not be guaranteed for all
                 backends).

        Generates:
            obj_id: object ids.

        """
        ...



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
import logging
import math
import os
import shlex
import socket
import stat
import subprocess
import time
from types import TracebackType
from typing import Callable, Dict, Iterable, Optional, Tuple, Type

from systemd.daemon import notify
from typing_extensions import Literal

from swh.perfecthash import Shard, ShardCreator

from .sharedbase import SharedBase
from .sleep import sleep_exponential
from .throttler import Throttler

logger = logging.getLogger(__name__)

# This would be used for image features that are not supported by the kernel RBD
# driver, e.g. exclusive-lock, object-map and fast-diff for kernels < 5.3
DEFAULT_IMAGE_FEATURES_UNSUPPORTED: Tuple[str, ...] = ()


class ShardNotMapped(Exception):
    pass


class Pool(object):
    """Manage a Ceph RBD pool for Winery shards.

    Arguments:
      shard_max_size: max size of shard contents
      rbd_use_sudo: whether to use sudo for rbd commands
      rbd_pool_name: name of the pool used for RBD images (metadata)
      rbd_data_pool_name: name of the pool used for RBD images (data)
      rbd_image_features_unsupported: features not supported by the kernel
        mounting the rbd images
      rbd_map_options: options to pass to ``rbd device map``, e.g.
        ``ms_mode=prefer-secure`` to connect to a ceph cluster with encryption
        enabled
    """

    def __init__(
        self,
        shard_max_size: int,
        rbd_use_sudo: bool = True,
        rbd_pool_name: str = "shards",
        rbd_data_pool_name: Optional[str] = None,
        rbd_image_features_unsupported: Tuple[
            str, ...
        ] = DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
        rbd_map_options: str = "",
    ) -> None:
        self.use_sudo = rbd_use_sudo
        self.pool_name = rbd_pool_name
        self.data_pool_name = rbd_data_pool_name or f"{self.pool_name}-data"
        self.features_unsupported = rbd_image_features_unsupported
        self.map_options = rbd_map_options
        self.image_size = math.ceil((shard_max_size * 2) / (1024 * 1024))

    POOL_CONFIG: Tuple[str, ...] = (
        "shard_max_size",
        "rbd_use_sudo",
        "rbd_pool_name",
        "rbd_data_pool_name",
        "rbd_image_features_unsupported",
        "rbd_map_options",
    )

    @classmethod
    def from_kwargs(cls, **kwargs) -> "Pool":
        """Create a Pool from a set of arbitrary keyword arguments"""
        return cls(**{k: kwargs[k] for k in cls.POOL_CONFIG if k in kwargs})

    def run(self, *cmd: str) -> Iterable[str]:
        """Run the given command, and return its output as lines.

        Return: the standard output of the run command

        Raises: CalledProcessError if the command doesn't exit with exit code 0.
        """

        sudo = ("sudo",) if self.use_sudo else ()
        cmd = sudo + cmd

        logger.debug(" ".join(repr(item) if " " in item else item for item in cmd))
        result = subprocess.check_output(cmd, encoding="utf-8", stderr=subprocess.PIPE)

        return result.splitlines()

    def rbd(self, *arguments: str) -> Iterable[str]:
        """Run rbd with the given arguments"""

        return self.run("rbd", f"--pool={self.pool_name}", *arguments)

    def image_exists(self, image: str):
        try:
            self.rbd("info", image)
        except subprocess.CalledProcessError:
            return False
        else:
            return True

    def image_mapped(self, image: str) -> Optional[Literal["ro", "rw"]]:
        """Check whether the image is already mapped, read-only or read-write"""
        try:
            image_stat = os.stat(self.image_path(image))
        except FileNotFoundError:
            return None
        return "rw" if (image_stat.st_mode & 0o222) != 0 else "ro"

    def image_list(self):
        try:
            images = self.rbd("ls")
        except subprocess.CalledProcessError as exc:
            if exc.returncode == 2 and "No such file or directory" in exc.stderr:
                return []
            else:
                raise
        return [image.strip() for image in images]

    def image_path(self, image: str) -> str:
        return f"/dev/rbd/{self.pool_name}/{image}"

    def image_create(self, image: str):
        self.rbd(
            "create",
            f"--size={self.image_size}",
            f"--data-pool={self.data_pool_name}",
            image,
        )
        if self.features_unsupported:
            self.rbd(
                "feature",
                "disable",
                f"{self.pool_name}/{image}",
                *self.features_unsupported,
            )
        self.image_map(image, "rw")

    def image_map(self, image: str, options: str):
        self.rbd(
            "device",
            "map",
            "-o",
            f"{options},{self.map_options}" if self.map_options else options,
            image,
        )

    def image_remap_ro(self, image: str):
        self.image_unmap(image)
        self.image_map(image, "ro")

    def image_unmap(self, image: str):
        if os.path.exists(self.image_path(image)):
            try:
                self.rbd("device", "unmap", self.image_path(image))
            except subprocess.CalledProcessError as exc:
                if exc.returncode == 22 and "Invalid argument" in exc.stderr:
                    logger.warning(
                        "Image %s already unmapped? stderr: %s", image, exc.stderr
                    )
                else:
                    raise

    def manage_images(
        self,
        base_dsn: str,
        manage_rw_images: bool,
        wait_for_image: Callable[[int], None],
        stop_running: Callable[[], bool],
        application_name: Optional[str] = None,
    ) -> None:
        """Manage RBD image creation and mapping automatically.

        Arguments:
          base_dsn: the DSN of the connection to the SharedBase
          manage_rw_images: whether RW images should be created and mapped
          wait_for_image: function which is called at each loop iteration, with
            an attempt number, if no images had to be mapped recently
          stop_running: callback that returns True when the manager should stop running
          application_name: the application name sent to PostgreSQL
        """
        application_name = application_name or "Winery RBD image manager"
        base = SharedBase(base_dsn=base_dsn, application_name=application_name)

        mapped_images: Dict[str, Literal["ro", "rw"]] = {}

        attempt = 0
        notified_systemd = False
        while not stop_running():
            did_something = False
            logger.debug("Listing shards")
            start = time.monotonic()
            shards = list(base.list_shards())
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    "Listed %d shards in %.02f seconds",
                    len(shards),
                    time.monotonic() - start,
                )
                logger.debug("Mapped images: %s", Counter(mapped_images.values()))

            notify(
                f"STATUS="
                "Enumerated {len(shards)} shards, "
                f"mapped {len(mapped_images)} images"
            )

            for shard_name, shard_state in shards:
                mapped_state = mapped_images.get(shard_name)
                if mapped_state == "ro":
                    continue
                elif shard_state.image_available:
                    check_mapped = self.image_mapped(shard_name)
                    if check_mapped == "ro":
                        logger.debug(
                            "Detected %s shard %s, already mapped read-only",
                            shard_state.name,
                            shard_name,
                        )
                    elif check_mapped == "rw":
                        logger.info(
                            "Detected %s shard %s, remapping read-only",
                            shard_state.name,
                            shard_name,
                        )
                        self.image_remap_ro(shard_name)
                        attempt = 0
                        while self.image_mapped(shard_name) != "ro":
                            attempt += 1
                            time.sleep(0.1)
                            if attempt % 100 == 0:
                                logger.warning(
                                    "Waiting for %s shard %s to be remapped "
                                    "read-only (for %ds)",
                                    shard_state.name,
                                    shard_name,
                                    attempt / 10,
                                )
                        base.record_shard_mapped(
                            name=shard_name, host=socket.gethostname()
                        )
                        did_something = True
                    else:
                        logger.debug(
                            "Detected %s shard %s, mapping read-only",
                            shard_state.name,
                            shard_name,
                        )
                        self.image_map(shard_name, options="ro")
                        base.record_shard_mapped(
                            name=shard_name, host=socket.gethostname()
                        )
                        did_something = True
                    mapped_images[shard_name] = "ro"
                elif manage_rw_images:
                    if os.path.exists(self.image_path(shard_name)):
                        # Image already mapped, nothing to do
                        pass
                    elif not self.image_exists(shard_name):
                        logger.info(
                            "Detected %s shard %s, creating RBD image",
                            shard_state.name,
                            shard_name,
                        )
                        self.image_create(shard_name)
                        did_something = True
                    else:
                        logger.warn(
                            "Detected %s shard %s and RBD image exists, mapping read-write",
                            shard_state.name,
                            shard_name,
                        )
                        self.image_map(shard_name, "rw")
                        did_something = True
                    # Now the shard is mapped
                    mapped_images[shard_name] = "rw"
                else:
                    logger.debug("%s shard %s, skipping", shard_state.name, shard_name)

            if not notified_systemd:
                # The first iteration has happened, all known shards should be ready
                notify("READY=1")
                notified_systemd = True

            if did_something:
                attempt = 0
            else:
                # Sleep using the current value
                wait_for_image(attempt)
                attempt += 1


class ROShard:
    def __init__(self, name, **kwargs):
        self.pool = Pool.from_kwargs(**kwargs)
        image_status = self.pool.image_mapped(name)

        if image_status != "ro":
            raise ShardNotMapped(
                f"RBD image for {name} isn't mapped{' read-only' if image_status=='rw' else ''}"
            )

        self.throttler = Throttler(**kwargs)
        self.name = name
        self.path = self.pool.image_path(self.name)
        self.shard = None
        self.open()
        logger.debug("ROShard %s: loaded", self.name)

    def open(self):
        try:
            self.shard = Shard(self.path)
        except FileNotFoundError:
            raise ShardNotMapped(f"RBD image for {self.name} not found at {self.path}")

    def get(self, key):
        if not self.shard:
            self.open()

        return self.throttler.throttle_get(self.shard.lookup, key)

    def close(self):
        if shard := getattr(self, "shard", None):
            shard.close()
        self.shard = None

    def __del__(self):
        self.close()

    @staticmethod
    def delete(pool, shard_name, obj_id):
        image_status = pool.image_mapped(shard_name)
        if image_status == "ro":
            raise PermissionError(
                f"Cannot delete object from {shard_name}, mapped read-only"
            )
        if not image_status:
            pool.image_map(shard_name, options="rw")
        Shard.delete(pool.image_path(shard_name), obj_id)


class ROShardCreator:
    """Helper for Read-Only shard creation.

    Arguments:
      name: Name of the shard to be initialized
      count: Number of objects to provision in the shard
      rbd_create_images: whether the ROShardCreator should create the rbd
        image, or delegate to the rbd_shard_manager
      rbd_wait_for_image: function called when waiting for a shard to be mapped
      shard_max_size: the size of the shard, passed to :class:`Pool`
      rbd_*: other RBD-related :class:`Pool` arguments
      throttle_*: :class:`Throttler` arguments
    """

    def __init__(
        self,
        name: str,
        count: int,
        rbd_create_images: bool = True,
        rbd_wait_for_image: Callable[[int], None] = sleep_exponential(
            min_duration=5,
            factor=2,
            max_duration=60,
            message="Waiting for RBD image mapping",
        ),
        **kwargs,
    ):
        self.pool = Pool.from_kwargs(**kwargs)
        self.throttler = Throttler(**kwargs)
        self.name = name
        self.count = count
        self.path = self.pool.image_path(self.name)
        self.rbd_create_images = rbd_create_images
        self.rbd_wait_for_image = rbd_wait_for_image

    def __enter__(self) -> "ROShardCreator":
        if self.rbd_create_images:
            self.pool.image_create(self.name)
        else:
            attempt = 0
            while not os.path.exists(self.path):
                self.rbd_wait_for_image(attempt)
                attempt += 1

        self.zero_image_if_needed()

        self.shard = ShardCreator(self.path, self.count)
        logger.debug("ROShard %s: created", self.name)
        self.shard.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.shard.__exit__(exc_type, exc_val, exc_tb)
        if self.rbd_create_images and not exc_type:
            self.pool.image_remap_ro(self.name)

    def zero_image_if_needed(self):
        """Check whether the image is empty, and zero it out if it's not.

        We really check only the first 1kB, as we assume that the SWHShard
        marker will have been written at the beginning of the image under all
        circumstances if the RO Shard creation has been interrupted.
        """
        with open(self.path, "rb") as f:
            start = f.read(1024)
            if set(start) == {0}:
                return

        logger.warning("RO Shard %s isn't empty, cleaning it up", self.path)
        st = os.stat(self.path)
        if stat.S_ISBLK(st.st_mode):
            # Block device, use DISCARD
            command = ["/usr/sbin/blkdiscard", self.path]
        else:
            # Regular file, use fallocate --punch-hole
            command = [
                "/usr/bin/fallocate",
                "--punch-hole",
                "-l",
                str(st.st_size),
                self.path,
            ]
        try:
            subprocess.run(command, check=True, capture_output=True)
        except subprocess.CalledProcessError:
            logger.warning("%s failed:", shlex.join(command), self.path, exc_info=True)

    def add(self, content, obj_id):
        return self.throttler.throttle_add(self.shard.write, obj_id, content)



# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy

import pytest
import yaml

from swh.objstorage.api.server import load_and_check_config


def prepare_config_file(tmpdir, content, name="config.yml"):
    """Prepare configuration file in `$tmpdir/name` with content `content`.

    Args:
        tmpdir (LocalPath): root directory
        content (str/dict): Content of the file either as string or as a dict.
                            If a dict, converts the dict into a yaml string.
        name (str): configuration filename

    Returns
        path (str) of the configuration file prepared.

    """
    config_path = tmpdir / name
    if isinstance(content, dict):  # convert if needed
        content = yaml.dump(content)
    config_path.write_text(content, encoding="utf-8")
    # pytest on python3.5 does not support LocalPath manipulation, so
    # convert path to string
    return str(config_path)


def test_load_and_check_config_no_configuration():
    """Inexistent configuration files raises"""
    with pytest.raises(EnvironmentError, match="Configuration file must be defined"):
        load_and_check_config(None)

    config_path = "/indexer/inexistent/config.yml"
    with pytest.raises(FileNotFoundError, match=f"{config_path} does not exist"):
        load_and_check_config(config_path)


def test_load_and_check_config_invalid_configuration_toplevel(tmpdir):
    """Invalid configuration raises"""
    config = {"something": "useless"}
    config_path = prepare_config_file(tmpdir, content=config)
    with pytest.raises(KeyError, match="missing objstorage config entry"):
        load_and_check_config(config_path)


def test_load_and_check_config_invalid_configuration(tmpdir):
    """Invalid configuration raises"""
    config_path = prepare_config_file(
        tmpdir, content={"objstorage": {"something": "useless"}}
    )
    with pytest.raises(KeyError, match="missing cls config entry"):
        load_and_check_config(config_path)


def test_load_and_check_config_invalid_configuration_level2(tmpdir):
    """Invalid configuration at 2nd level raises"""
    config = {
        "objstorage": {
            "cls": "pathslicing",
            "root": "root",
            "slicing": "slicing",
            "client_max_size": "10",
        }
    }
    for key in ("root", "slicing"):
        c = copy.deepcopy(config)
        c["objstorage"].pop(key)
        config_path = prepare_config_file(tmpdir, c)
        with pytest.raises(KeyError, match=f"missing {key} config entry"):
            load_and_check_config(config_path)


@pytest.mark.parametrize(
    "config",
    [
        pytest.param(
            {
                "objstorage": {
                    "cls": "pathslicing",
                    "root": "root",
                    "slicing": "slicing",
                }
            },
            id="pathslicing-bw-compat",
        ),
        pytest.param(
            {
                "objstorage": {
                    "cls": "pathslicing",
                    "root": "root",
                    "slicing": "slicing",
                }
            },
            id="pathslicing",
        ),
        pytest.param(
            {"client_max_size": "10", "objstorage": {"cls": "memory"}},
            id="empty-args-bw-compat",
        ),
        pytest.param(
            {"client_max_size": "10", "objstorage": {"cls": "memory"}}, id="empty-args"
        ),
        pytest.param(
            {
                "objstorage": {
                    "cls": "noop",
                }
            },
            id="noop",
        ),
    ],
)
def test_load_and_check_config(tmpdir, config):
    """pathslicing configuration fine loads ok"""
    config_path = prepare_config_file(tmpdir, config)
    cfg = load_and_check_config(config_path)
    assert cfg == config



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, Iterator, Union

from swh.objstorage.exc import ReadOnlyObjStorageError
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import CompositeObjId, ObjStorageInterface
from swh.objstorage.objstorage import ObjStorage


class ReadOnlyProxyObjStorage(ObjStorage):
    """Filter that disable write operation of the storage.

    Writes will always succeed without doing any actual write operations.
    """

    name: str = "read-only"

    def __init__(self, storage: Union[ObjStorageInterface, Dict], **kwargs):
        super().__init__(**kwargs)
        self.storage: ObjStorageInterface = (
            get_objstorage(**storage) if isinstance(storage, dict) else storage
        )

    def __contains__(self, *args, **kwargs):
        return self.storage.__contains__(*args, **kwargs)

    def __iter__(self) -> Iterator[CompositeObjId]:
        """Iterates over the content of each storages

        Warning: The `__iter__` methods frequently have bad performance. You
        almost certainly don't want to use this method in production as the
        wrapped storage may cause performance issues.
        """
        return self.storage.__iter__()

    def __len__(self):
        """Compute the number of objects in the current object storage.

        Warning: performance issue in `__iter__` also applies here.

        Returns:
            number of objects contained in the storage.
        """
        return self.storage.__len__()

    def get(self, obj_id, *args, **kwargs):
        return self.storage.get(obj_id, *args, **kwargs)

    def check(self, obj_id, *args, **kwargs):
        return self.storage.check(obj_id, *args, **kwargs)

    def check_config(self, *, check_write):
        if check_write:
            return False
        return self.storage.check_config(check_write=False)

    def add(self, *args, **kwargs):
        raise ReadOnlyObjStorageError("add")

    def restore(self, *args, **kwargs):
        raise ReadOnlyObjStorageError("restore")

    def delete(self, *args, **kwargs):
        raise ReadOnlyObjStorageError("dalete")



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import ExitStack
from enum import Enum
import logging
from types import TracebackType
from typing import Iterator, Optional, Set, Tuple, Type
import uuid

import psycopg
import psycopg.errors

from .database import Database

WRITER_UUID = uuid.uuid4()

logger = logging.getLogger(__name__)


class ShardState(Enum):
    STANDBY = "standby"
    WRITING = "writing"
    FULL = "full"
    PACKING = "packing"
    PACKED = "packed"
    CLEANING = "cleaning"
    READONLY = "readonly"

    @property
    def locked(self):
        return self not in {self.STANDBY, self.FULL, self.PACKED, self.READONLY}

    @property
    def image_available(self):
        return self in {self.PACKED, self.CLEANING, self.READONLY}

    @property
    def readonly(self):
        return self in {self.CLEANING, self.READONLY}


class SignatureState(Enum):
    INFLIGHT = "inflight"
    PRESENT = "present"
    DELETED = "deleted"


class TemporaryShardLocker:
    """Opportunistically lock a shard, and provide a context manager to unlock
    the shard if an operation fails.

    Use this through the :meth:`SharedBase.maybe_lock_one_shard` method.
    """

    def __init__(
        self,
        base: "SharedBase",
        current_state: ShardState,
        new_state: ShardState,
        min_mapped_hosts: int = 0,
    ) -> None:
        self.base = base
        self.previous_state = current_state
        self.name: Optional[str] = None
        self.id: Optional[int] = None
        locked = self.base.lock_one_shard(
            current_state=current_state,
            new_state=new_state,
            min_mapped_hosts=min_mapped_hosts,
        )
        if locked:
            self.name, self.id = locked

    def __bool__(self) -> bool:
        return self.name is not None

    def __enter__(self) -> "TemporaryShardLocker":
        if not self:
            raise ValueError("No shard was locked")
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        traceback: Optional[TracebackType],
    ) -> None:
        """If an exception has been raised, restore the shard to its previous state"""
        if not self:
            return

        if not exc_type:
            return

        try:
            self.base.set_shard_state(
                name=self.name, new_state=self.previous_state, check_locker=True
            )
        except Exception:
            logger.warning("Could not unlock shard %s:", self.name, exc_info=True)


class SharedBase(Database):
    current_version: int = 1

    def __init__(
        self, base_dsn: str, application_name: str = "SWH Winery SharedBase", **kwargs
    ) -> None:
        super().__init__(
            dsn=base_dsn,
            application_name=application_name,
        )
        self._locked_shard: Optional[Tuple[str, int]] = None

        logger.debug("SharedBase %s: instantiated", WRITER_UUID)

    @property
    def locked_shard(self) -> str:
        self.set_locked_shard()

        assert self._locked_shard, "failed to lock a shard"
        return self._locked_shard[0]

    @property
    def locked_shard_id(self) -> int:
        self.set_locked_shard()

        assert self._locked_shard, "failed to lock a shard"
        return self._locked_shard[1]

    def set_locked_shard(self) -> None:
        if self._locked_shard is not None:
            return

        locked = self.lock_one_shard(
            current_state=ShardState.STANDBY, new_state=ShardState.WRITING
        )
        if locked is not None:
            self._locked_shard = locked
        else:
            if logger.isEnabledFor(logging.DEBUG):
                import traceback

                stack = traceback.extract_stack()
                for item in stack[::-1]:
                    if item.filename != __file__:
                        logger.debug(
                            "Creating new shard from file %s, line %d, function %s: %s",
                            item.filename,
                            item.lineno,
                            item.name,
                            item.line,
                        )
                        break
            self._locked_shard = self.create_shard(new_state=ShardState.WRITING)

        return

    def lock_one_shard(
        self,
        current_state: ShardState,
        new_state: ShardState,
        min_mapped_hosts: int = 0,
    ) -> Optional[Tuple[str, int]]:
        """Lock one shard in `current_state`, putting it into `new_state`. Only
        lock a shard if it has more than `min_mapped_hosts` hosts that have
        registered as having mapped the shard.
        """

        if not new_state.locked:
            raise ValueError(f"{new_state} is not a locked state")

        with self.pool.connection() as db, db.transaction():
            with db.cursor() as c:
                # run the next two statements in a transaction
                c.execute(
                    """\
                    SELECT name
                    FROM shards
                    WHERE
                      state = %s
                      AND COALESCE(ARRAY_LENGTH(mapped_on_hosts_when_packed, 1), 0) >= %s
                    LIMIT 1
                    FOR UPDATE SKIP LOCKED
                    """,
                    (current_state.value, min_mapped_hosts),
                )
                result = c.fetchone()
                if result is None:
                    return None
                shard_name = result[0]
                try:
                    c.execute(
                        """\
                        UPDATE shards
                        SET state = %s, locker_ts = now(), locker = %s
                        WHERE name = %s
                        RETURNING name, id
                        """,
                        (
                            new_state.value,
                            WRITER_UUID,
                            shard_name,
                        ),
                    )
                except psycopg.Error:
                    logger.exception(
                        "SharedBase %s: shard %s failed to lock",
                        WRITER_UUID,
                        shard_name,
                    )
                    return None
                else:
                    logger.debug(
                        "SharedBase %s: shard %s locked", WRITER_UUID, shard_name
                    )
                    return c.fetchone()

    def maybe_lock_one_shard(
        self,
        current_state: ShardState,
        new_state: ShardState,
        min_mapped_hosts: int = 0,
    ) -> TemporaryShardLocker:
        """Opportunistically lock a shard, and, if a shard was locked, provide a
        context manager to rollback the locking on failure.

        Example::

           locked = base.maybe_lock_one_shard(
               current_state=ShardState.FULL,
               new_state=ShardState.PACKING,
           )
           if not locked:
               wait_a_minute()
               return

           with locked:
               do_something_with_locked_shard(locked.name)

        If ``do_something_with_locked_shard`` fails, the shard will be moved
        back to the ``current_state`` on exit.
        """
        return TemporaryShardLocker(
            base=self,
            current_state=current_state,
            new_state=new_state,
            min_mapped_hosts=min_mapped_hosts,
        )

    def set_shard_state(
        self,
        new_state: ShardState,
        set_locker: bool = False,
        check_locker: bool = False,
        name: Optional[str] = None,
        db: Optional[psycopg.Connection] = None,
    ):
        if not name:
            if not self._locked_shard:
                raise ValueError("Can't set shard state, no shard specified or locked")
            name = self._locked_shard[0]

        reset_locked_shard = self._locked_shard and self._locked_shard[0] == name

        with ExitStack() as stack:
            if not db:
                db = stack.enter_context(self.pool.connection())

            assert isinstance(db, psycopg.Connection)

            c = db.execute(
                """\
                UPDATE shards
                SET
                  locker = %s,
                  locker_ts = (CASE WHEN %s THEN NOW() ELSE NULL END),
                  state = %s
                WHERE name = %s AND (CASE WHEN %s THEN locker = %s ELSE TRUE END)
                """,
                (
                    WRITER_UUID if set_locker else None,
                    set_locker,
                    new_state.value,
                    name,
                    check_locker,
                    WRITER_UUID,
                ),
            )
            affected = c.rowcount
            if affected != 1:
                raise ValueError(
                    "set_shard_state(%s) affected %s rows, expected 1"
                    % (name, affected)
                )

            logger.debug(
                "SharedBase %s: shard %s moved into state %s",
                WRITER_UUID,
                name,
                new_state,
            )

        if reset_locked_shard and not new_state.locked:
            self._locked_shard = None

    def create_shard(self, new_state: ShardState) -> Tuple[str, int]:
        name = uuid.uuid4().hex
        #
        # ensure the first character is not a number so it can be used as a
        # database name.
        #
        name = "i" + name[1:]
        with self.pool.connection() as db, db.cursor() as c:
            c.execute(
                """\
                INSERT INTO shards
                  (name, state, locker, locker_ts)
                VALUES
                  (%s, %s, %s, NOW())
                RETURNING name, id""",
                (name, new_state.value, WRITER_UUID),
            )
            res = c.fetchone()
            if res is None:
                raise RuntimeError(
                    f"Writer {WRITER_UUID} failed to create shard with name {name}"
                )

            logger.debug(
                "SharedBase %s: shard %s created (and locked)", WRITER_UUID, res[0]
            )
            return res

    def shard_packing_starts(self, name: str):
        with self.pool.connection() as db, db.transaction():
            with db.cursor() as c:
                c.execute(
                    "SELECT state FROM shards WHERE name=%s FOR UPDATE SKIP LOCKED",
                    (name,),
                )
                returned = c.fetchone()
                if not returned:
                    raise ValueError("Could not get shard state for %s" % name)
                state = ShardState(returned[0])
                if state != ShardState.FULL:
                    raise ValueError(
                        "Cannot pack shard in state %s, expected ShardState.FULL"
                        % state
                    )

                logger.debug(
                    "SharedBase %s: shard %s starts packing", WRITER_UUID, name
                )
                self.set_shard_state(
                    name=name,
                    new_state=ShardState.PACKING,
                    set_locker=True,
                    check_locker=False,
                    db=db,
                )

    def shard_packing_ends(self, name):
        with self.pool.connection() as db, db.transaction():
            with db.cursor() as c:
                c.execute(
                    "SELECT state FROM shards WHERE name=%s FOR UPDATE SKIP LOCKED",
                    (name,),
                )
                returned = c.fetchone()
                if not returned:
                    raise ValueError("Could not get shard state for %s" % name)
                state = ShardState(returned[0])
                if state != ShardState.PACKING:
                    raise ValueError(
                        "Cannot finalize packing for shard in state %s,"
                        " expected ShardState.PACKING" % state
                    )

                logger.debug("SharedBase %s: shard %s done packing", WRITER_UUID, name)
                self.set_shard_state(
                    name=name,
                    new_state=ShardState.PACKED,
                    set_locker=False,
                    check_locker=False,
                    db=db,
                )

    def get_shard_info(self, id: int) -> Optional[Tuple[str, ShardState]]:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute("SELECT name, state FROM shards WHERE id = %s", (id,))
            row = c.fetchone()
            if not row:
                return None
            return (row[0], ShardState(row[1]))

    def get_shard_state(self, name: str) -> Optional[ShardState]:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute("SELECT state FROM shards WHERE name = %s", (name,))
            row = c.fetchone()
            if not row:
                return None
            return ShardState(row[0])

    def list_shards(self) -> Iterator[Tuple[str, ShardState]]:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute("SELECT name, state FROM shards")
            for row in c:
                yield row[0], ShardState(row[1])

    def count_objects(self, name: Optional[str] = None) -> Optional[int]:
        if not name:
            if not self._locked_shard:
                raise ValueError("Can't count objects, no shard specified or locked")
            name = self._locked_shard[0]

        with self.pool.connection() as db:
            c = db.execute(
                """SELECT shard, COUNT(*)
                         FROM signature2shard
                         WHERE state = 'present'
                           AND shard = (SELECT id FROM shards WHERE name = %s)
                         GROUP BY shard
                """,
                (name,),
            )
            row = c.fetchone()
            if not row:
                return 0
            return row[1]

    def record_shard_mapped(self, host: str, name: str) -> Set[str]:
        with self.pool.connection() as db, db.transaction():
            with db.cursor() as c:
                c.execute(
                    """SELECT mapped_on_hosts_when_packed
                       FROM shards
                       WHERE name = %s
                       FOR UPDATE SKIP LOCKED""",
                    (name,),
                )
                row = c.fetchone()
                if not row:
                    raise ValueError("Can't update shard %s" % name)
                hosts = set(row[0])
                if host not in hosts:
                    hosts.add(host)
                    c.execute(
                        """UPDATE shards
                           SET mapped_on_hosts_when_packed = %s
                           WHERE name = %s""",
                        (list(hosts), name),
                    )
                return hosts

    def contains(self, obj_id) -> Optional[int]:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute(
                "SELECT shard FROM signature2shard WHERE "
                "signature = %s AND state = 'present'",
                (obj_id,),
            )
            row = c.fetchone()
            if not row:
                return None
            return row[0]

    def get(self, obj_id) -> Optional[Tuple[str, ShardState]]:
        id = self.contains(obj_id)
        if id is None:
            return None
        return self.get_shard_info(id)

    def record_new_obj_id(self, db, obj_id) -> Optional[int]:
        db.execute(
            "INSERT INTO signature2shard (signature, shard, state) "
            "VALUES (%s, %s, 'present') ON CONFLICT (signature) DO NOTHING",
            (obj_id, self.locked_shard_id),
        )
        cur = db.execute(
            "SELECT shard FROM signature2shard WHERE signature = %s", (obj_id,)
        )
        return cur.fetchone()[0]

    def delete(self, obj_id):
        with self.pool.connection() as db:
            db.execute(
                "UPDATE signature2shard SET state = 'deleted' WHERE signature = %s",
                (obj_id,),
            )

    def deleted_objects(self) -> Iterator[Tuple[bytes, str, ShardState]]:
        with self.pool.connection() as db:
            cur = db.execute(
                """SELECT signature, shards.name, shards.state
               FROM signature2shard objs, shards
               WHERE objs.state = 'deleted'
                 AND shards.id = objs.shard
               """
            )
            for signature, name, state in cur.fetchall():
                yield bytes(signature), name, ShardState(state)

    def clean_deleted_object(self, obj_id) -> None:
        with self.pool.connection() as db:
            db.execute("DELETE FROM signature2shard WHERE signature = %s", (obj_id,))



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, Iterator, Optional

import msgpack

from swh.core.api import RPCClient
from swh.objstorage.constants import DEFAULT_LIMIT
from swh.objstorage.exc import (
    Error,
    NoBackendsLeftError,
    ObjCorruptedError,
    ObjNotFoundError,
    ObjStorageAPIError,
)
from swh.objstorage.interface import CompositeObjId, ObjId, ObjStorageInterface
from swh.objstorage.objstorage import objid_to_default_hex, timed


class RemoteObjStorage(RPCClient):
    """Proxy to a remote object storage.

    This class allows to connect to an object storage server via
    http protocol.

    Attributes:
        url (string): The url of the server to connect. Must end
            with a '/'
        session: The session to send requests.

    """

    api_exception = ObjStorageAPIError
    reraise_exceptions = [
        ObjNotFoundError,
        Error,
        ObjCorruptedError,
        NoBackendsLeftError,
    ]
    backend_class = ObjStorageInterface
    name: str = "remote"

    def restore(self: ObjStorageInterface, content: bytes, obj_id: ObjId) -> None:
        return self.add(content, obj_id, check_presence=False)

    def __iter__(self) -> Iterator[CompositeObjId]:
        yield from self.list_content()

    def list_content(
        self,
        last_obj_id: Optional[ObjId] = None,
        limit: Optional[int] = DEFAULT_LIMIT,
    ) -> Iterator[CompositeObjId]:
        params: Dict[str, Any] = {}
        if limit:
            params["limit"] = limit
        if last_obj_id:
            params["last_obj_id"] = objid_to_default_hex(last_obj_id)
        response = self.raw_verb(
            "get",
            "content",
            headers={"accept": "application/x-msgpack"},
            params=params,
            stream=True,
        )
        yield from msgpack.Unpacker(response.raw, raw=False)


# XXX Maybe there is a better way of doing this, but according the automagic
# way this class is built (via the MetaRPCClient metaclass), one cannot easily
# just overload the methods in the class definition.
RemoteObjStorage.get = timed(RemoteObjStorage.get)  # type: ignore
RemoteObjStorage.add = timed(RemoteObjStorage.add)  # type: ignore
RemoteObjStorage.__contains__ = timed(RemoteObjStorage.__contains__)  # type: ignore



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections.abc import Iterator

from swh.objstorage.backends.generator import Randomizer
from swh.objstorage.factory import get_objstorage


def test_random_generator_objstorage():
    sto = get_objstorage("random")
    assert sto

    blobs = [sto.get(None) for i in range(100)]
    lengths = [len(x) for x in blobs]
    assert max(lengths) <= 55056238


def test_random_generator_objstorage_list_content():
    sto = get_objstorage("random", total=100)
    assert isinstance(sto.list_content(), Iterator)

    assert list(sto.list_content()) == [{"sha1": b"%d" % i} for i in range(1, 101)]
    assert list(sto.list_content(limit=10)) == [
        {"sha1": b"%d" % i} for i in range(1, 11)
    ]
    assert list(sto.list_content(last_obj_id=b"10", limit=10)) == [
        {"sha1": b"%d" % i} for i in range(11, 21)
    ]


def test_random_generator_objstorage_total():
    sto = get_objstorage("random", total=5)
    assert len([x for x in sto]) == 5


def test_random_generator_objstorage_size():
    sto = get_objstorage("random", filesize=10)
    for i in range(10):
        assert len(sto.get(None)) == 10


def test_randomizer_read_size():
    r = Randomizer()
    for i in range(-2, 3):
        size = r.size + i
        assert len(r.read(size)) == size
        if i >= 0:
            # Randomizer buffer should have been increased
            assert r.size > size



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio

from swh.objstorage.interface import COMPOSITE_OBJID_KEYS, ObjId


def call_async(f, *args):
    """Calls an async coroutine from a synchronous function."""
    loop = asyncio.new_event_loop()
    try:
        return loop.run_until_complete(f(*args))
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()


def format_obj_id(obj_id: ObjId) -> str:
    if isinstance(obj_id, bytes):
        obj_id = {"sha1": obj_id}

    return ";".join(
        (
            "%s:%s" % (algo, obj_id[algo].hex())
            for algo in sorted(COMPOSITE_OBJID_KEYS)
            if algo in obj_id
        )
    )



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import pytest

from swh.core.api.tests.server_testing import ServerTestFixture
from swh.objstorage.api.server import app
from swh.objstorage.factory import get_objstorage
from swh.objstorage.tests.objstorage_testing import ObjStorageTestFixture


class TestRemoteObjStorage(ServerTestFixture, ObjStorageTestFixture):
    """Test the remote archive API."""

    @pytest.fixture(autouse=True)
    def objstorage(self, tmpdir):
        self.config = {
            "objstorage": {
                "cls": "pathslicing",
                "root": str(tmpdir),
                "slicing": "0:1/0:5",
                "allow_delete": True,
            },
            "client_max_size": 8 * 1024 * 1024,
        }

        self.app = app
        self.start_server()
        self.storage = get_objstorage("remote", url=self.url())
        yield
        self.stop_server()

    @pytest.mark.skip("makes no sense to test this for the remote api")
    def test_delete_not_allowed(self):
        pass

    @pytest.mark.skip("makes no sense to test this for the remote api")
    def test_delete_not_allowed_by_default(self):
        pass



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import atexit
import logging
import os
from pathlib import Path
from subprocess import CalledProcessError
from typing import Iterable, List, Optional, Tuple

from swh.objstorage.backends.winery.roshard import (
    DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
    Pool,
)

logger = logging.getLogger(__name__)


DEFAULT_ERASURE_CODE_PROFILE = {
    "name": "winery-test-profile",
    "k": "4",
    "m": "2",
    "crush-failure-domain": "host",
}
DEFAULT_DATA_POOL_SETTINGS = {
    "allow_ec_overwrites": "true",
    "pg_autoscale_mode": "off",
    "pg_num": 128,
}


class PoolHelper(Pool):
    def __init__(
        self,
        shard_max_size: int,
        rbd_use_sudo: bool = True,
        rbd_pool_name: str = "test-shards",
        rbd_data_pool_name: Optional[str] = None,
        rbd_image_features_unsupported: Tuple[
            str, ...
        ] = DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
        rbd_map_options: str = "",
        rbd_erasure_code_profile=None,
        rbd_data_pool_settings=None,
    ):
        super().__init__(
            shard_max_size=shard_max_size,
            rbd_use_sudo=rbd_use_sudo,
            rbd_pool_name=rbd_pool_name,
            rbd_data_pool_name=rbd_data_pool_name,
            rbd_image_features_unsupported=rbd_image_features_unsupported,
            rbd_map_options=rbd_map_options,
        )
        self.erasure_code_profile_settings = (
            rbd_erasure_code_profile or DEFAULT_ERASURE_CODE_PROFILE.copy()
        )
        self.erasure_code_profile = self.erasure_code_profile_settings.pop(
            "name", DEFAULT_ERASURE_CODE_PROFILE["name"]
        )

        self.data_pool_settings = (
            rbd_data_pool_settings or DEFAULT_DATA_POOL_SETTINGS.copy()
        )
        self.data_pool_pg_num = self.data_pool_settings.pop(
            "pg_num", DEFAULT_DATA_POOL_SETTINGS["pg_num"]
        )

    POOL_CONFIG = Pool.POOL_CONFIG + (
        "rbd_erasure_code_profile",
        "rbd_data_pool_settings",
    )

    def ceph(self, *arguments) -> Iterable[str]:
        """Run ceph with the given arguments"""

        return self.run("ceph", *arguments)

    def image_remove(self, image):
        self.image_unmap(image)
        self.rbd("remove", image)

    def images_remove(self):
        for image in self.image_list():
            try:
                self.image_remove(image)
            except CalledProcessError:
                logger.error(
                    "Could not remove image %s, we'll try again in an atexit handler...",
                    image,
                )
                atexit.register(self.image_remove, image)
                pass

    def remove(self):
        self.images_remove()
        self.pool_remove()

    def pool_remove(self):
        for pool in (self.pool_name, self.data_pool_name):
            self.ceph(
                "osd",
                "pool",
                "delete",
                pool,
                pool,
                "--yes-i-really-really-mean-it",
            )

    def pool_create(self):
        try:
            output = self.ceph(
                "osd",
                "erasure-code-profile",
                "get",
                self.erasure_code_profile,
            )
        except CalledProcessError:
            self.ceph(
                "osd",
                "erasure-code-profile",
                "set",
                self.erasure_code_profile,
                *(f"{k}={v}" for k, v in self.erasure_code_profile_settings.items()),
            )
        else:
            current_settings = dict(line.split("=", 1) for line in output)
            for k, v in self.erasure_code_profile_settings.items():
                if (current_setting := current_settings[k]) != str(v):
                    logger.warning(
                        "For erasure coding profile %s, setting %s=%s != requested %s",
                        self.erasure_code_profile,
                        k,
                        current_setting,
                        v,
                    )

        self.ceph(
            "osd",
            "pool",
            "create",
            self.data_pool_name,
            f"{self.data_pool_pg_num}",
            "erasure",
            self.erasure_code_profile,
        )

        for setting, value in self.data_pool_settings.items():
            self.ceph("osd", "pool", "set", self.data_pool_name, setting, value)

        self.ceph("osd", "pool", "create", self.pool_name)


class FileBackedPool(Pool):
    """File-backed pool for Winery shards mimicking a Ceph RBD pool.

    Unmapped images are represented by setting the file permission to 0o000.
    """

    base_directory: Optional[Path] = None

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        assert (
            FileBackedPool.base_directory is not None
        ), "set_base_directory() should have been called first"
        self.pool_dir = FileBackedPool.base_directory / self.pool_name
        self.pool_dir.mkdir(exist_ok=True)

    @classmethod
    def set_base_directory(cls, base_directory: Path) -> None:
        cls.base_directory = base_directory

    @classmethod
    def from_kwargs(cls, **kwargs) -> "Pool":
        """Create a Pool from a set of arbitrary keyword arguments"""
        return cls(**{k: kwargs[k] for k in Pool.POOL_CONFIG if k in kwargs})

    def run(self, *cmd: str) -> Iterable[str]:
        raise NotImplementedError

    def rbd(self, *arguments: str) -> Iterable[str]:
        raise NotImplementedError

    def image_exists(self, image: str) -> bool:
        return (self.pool_dir / image).is_file()

    def image_list(self) -> List[str]:
        return [entry.name for entry in self.pool_dir.iterdir() if entry.is_file()]

    def image_path(self, image: str) -> str:
        return str(self.pool_dir / image)

    def image_create(self, image: str) -> None:
        path = self.image_path(image)
        if os.path.exists(path):
            raise ValueError(f"Image {image} already exists")
        open(path, "w").close()
        os.truncate(path, self.image_size * 1024 * 1024)
        self.image_map(image, "rw")

    def image_map(self, image: str, options: str) -> None:
        if "ro" in options:
            os.chmod(self.image_path(image), 0o400)
        else:
            os.chmod(self.image_path(image), 0o600)

    def image_unmap(self, image: str) -> None:
        os.chmod(self.image_path(image), 0o000)

    def image_unmap_all(self) -> None:
        for entry in self.pool_dir.iterdir():
            if entry.is_file():
                entry.chmod(0o000)



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
from deprecated import deprecated


def _exc_message(base: str, args):
    message = base
    if args:
        message += f": {args[0]}"
    return message


class Error(Exception):
    def __str__(self):
        return _exc_message("Object storage error on object", self.args)


class ObjNotFoundError(Error):
    def __str__(self):
        return _exc_message("Object not found", self.args)


class ObjCorruptedError(Error):
    def __str__(self) -> str:
        return _exc_message("Object corrupted", self.args)


class NoBackendsLeftError(Error):
    def __str__(self) -> str:
        return _exc_message("No backends left to read from", self.args)


class ObjStorageAPIError(Exception):
    """Specific internal exception of an object storage (mainly connection)."""

    def __str__(self):
        return _exc_message(
            "An unexpected error occurred in the api backend", self.args
        )


class ReadOnlyObjStorageError(Error):
    def __init__(self, method, *args):
        super().__init__(*args)
        self.method = method

    def __str__(self):
        return f"This object storage is Read-Only: cannot use {self.method}"


ReadOnlyObjStorage = deprecated(
    version="v3.1.0",
    reason="This exception has been renamed as ReadOnlyObjStorageError",
)(ReadOnlyObjStorageError)


class NonIterableObjStorageError(Error):
    def __init__(self, method, *args):
        super().__init__(*args)
        self.method = method

    def __str__(self):
        return f"This object storage is not iterable: cannot use {self.method}"


NonIterableObjStorage = deprecated(
    version="v3.1.0",
    reason="This exception has been renamed as NonIterableObjStorageError",
)(NonIterableObjStorageError)



# Copyright (C) 2016-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import importlib
import warnings

from swh.objstorage.interface import ObjStorageInterface
from swh.objstorage.objstorage import ObjStorage

__all__ = ["get_objstorage", "ObjStorage"]


OBJSTORAGE_IMPLEMENTATIONS = {
    "pathslicing": "swh.objstorage.backends.pathslicing.PathSlicingObjStorage",
    "remote": "swh.objstorage.api.client.RemoteObjStorage",
    "memory": "swh.objstorage.backends.in_memory.InMemoryObjStorage",
    "seaweedfs": "swh.objstorage.backends.seaweedfs.objstorage.SeaweedFilerObjStorage",
    "random": "swh.objstorage.backends.generator.RandomGeneratorObjStorage",
    "http": "swh.objstorage.backends.http.HTTPReadOnlyObjStorage",
    "noop": "swh.objstorage.backends.noop.NoopObjStorage",
    "azure": "swh.objstorage.backends.azure.AzureCloudObjStorage",
    "azure-prefixed": "swh.objstorage.backends.azure.PrefixedAzureCloudObjStorage",
    "s3": "swh.objstorage.backends.libcloud.AwsCloudObjStorage",
    "swift": "swh.objstorage.backends.libcloud.OpenStackCloudObjStorage",
    "winery": "swh.objstorage.backends.winery.WineryObjStorage",
    # filters and proxies
    "multiplexer": "swh.objstorage.multiplexer.MultiplexerObjStorage",
    "read-only": "swh.objstorage.proxies.readonly.ReadOnlyProxyObjStorage",
    # deprecated factories
    "filtered": "_construct_filtered_objstorage",
}


def get_objstorage(cls: str, **kwargs) -> ObjStorageInterface:
    """Create an ObjStorage using the given implementation class.

    Args:
        cls: objstorage class unique key contained in the
            OBJSTORAGE_IMPLEMENTATIONS dict.
        kwargs: arguments for the required class of objstorage
                that must match exactly the one in the `__init__` method of the
                class.
    Returns:
        subclass of ObjStorage that match the given `storage_class` argument.
    Raises:
        ValueError: if the given storage class is not a valid objstorage
            key.
    """
    class_path = OBJSTORAGE_IMPLEMENTATIONS.get(cls)
    if class_path is None:
        raise ValueError(
            "Unknown storage class `%s`. Supported: %s"
            % (cls, ", ".join(OBJSTORAGE_IMPLEMENTATIONS))
        )

    if "." in class_path:
        (module_path, class_name) = class_path.rsplit(".", 1)
        try:
            module = importlib.import_module(module_path, package=__package__)
        except ImportError as e:
            raise ValueError(f"Storage class {cls} is not available: {e.args[0]}")
        ObjStorage = getattr(module, class_name)
    else:
        # used by (deprecated) filtered for which the value is a factory
        # function rather than a class
        ObjStorage = globals()[class_path]
    return ObjStorage(**kwargs)


def _construct_filtered_objstorage(storage_conf, filters_conf, **kwargs):
    if len(filters_conf) != 1 or filters_conf[0]["type"] != "readonly":
        raise ValueError("This legacy function only supports a single readonly filter")
    warnings.warn(
        "The 'filtered[type:readonly]' objstorage class has been deprecated, "
        "please use a 'read-only' proxy class instead.",
        DeprecationWarning,
    )

    return get_objstorage(cls="read-only", storage=get_objstorage(**storage_conf))



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import logging
import os
import time

logger = logging.getLogger(__name__)


class Stats:
    def __init__(self, d):
        if d is None:
            self._stats_active = False
            return

        self._stats_active = True
        if not os.path.exists(d):
            os.makedirs(d)
        self._stats = collections.Counter()
        self._stats_filename = f"{d}/{os.getpid()}.csv"
        self._stats_fd = open(self.stats_filename, "a")
        if self._stats_fd.tell() == 0:
            # We created the file, write the CSV header
            self._stats_fd.write(
                # time in seconds since epoch
                "time,"
                # total number of objects written at this point in time
                "object_write_count,"
                # total number of bytes written at this point in time
                "bytes_write,"
                # total number of objects read at this point in time
                "object_read_count,"
                # total number of bytes read at this point in time
                "bytes_read"
                "\n"
            )
        else:
            # The PID was recycled, print an empty stats line at the end of the file to mark it
            self._stats_print()

        self._stats_last_write = time.monotonic()
        self._stats_flush_interval = 5

    @property
    def stats_active(self):
        return self._stats_active

    @property
    def stats_filename(self):
        return self._stats_filename

    def __del__(self):
        if self.stats_active and not self._stats_fd.closed:
            self._stats_print()
            self._stats_fd.close()

    def _stats_print(self):
        ll = ",".join(
            str(self._stats[x])
            for x in [
                "object_write_count",
                "bytes_write",
                "object_read_count",
                "bytes_read",
            ]
        )
        self._stats_fd.write(f"{int(time.time())},{ll}\n")

    def _stats_maybe_print(self):
        now = time.monotonic()
        if now - self._stats_last_write > self._stats_flush_interval:
            self._stats_print()
            self._stats_last_write = now

    def stats_read(self, key, content):
        self._stats["object_read_count"] += 1
        self._stats["bytes_read"] += len(key) + len(content)
        self._stats_maybe_print()

    def stats_write(self, key, content):
        self._stats["object_write_count"] += 1
        self._stats["bytes_write"] += len(key) + len(content)
        self._stats_maybe_print()



# Copyright (C) 2016-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import contextlib
from datetime import datetime, timedelta
from itertools import product
import string
from typing import Iterable, Iterator, Mapping, Optional, Union
import warnings

from azure.core.exceptions import ResourceExistsError, ResourceNotFoundError
from azure.storage.blob import (
    BlobSasPermissions,
    ContainerClient,
    ContainerSasPermissions,
    generate_blob_sas,
    generate_container_sas,
)
from azure.storage.blob.aio import ContainerClient as AsyncContainerClient

from swh.model import hashutil
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.interface import CompositeObjId, ObjId
from swh.objstorage.objstorage import CompressionFormat, ObjStorage, timed
from swh.objstorage.utils import call_async


def get_container_url(
    account_name: str,
    account_key: str,
    container_name: str,
    access_policy: str = "read_only",
    expiry: timedelta = timedelta(days=365),
    container_url_template: str = (
        "https://{account_name}.blob.core.windows.net/{container_name}?{signature}"
    ),
    **kwargs,
) -> str:
    """Get the full url, for the given container on the given account, with a
    Shared Access Signature granting the specified access policy.

    Args:
      account_name: name of the storage account for which to generate the URL
      account_key: shared account key of the storage account used to generate the SAS
      container_name: name of the container for which to grant access in the storage
        account
      access_policy: one of ``read_only``, ``append_only``, ``full``
      expiry: the interval in the future with which the signature will expire

    Returns:
      the full URL of the container, with the shared access signature.
    """

    access_policies = {
        "read_only": ContainerSasPermissions(
            read=True, list=True, delete=False, write=False
        ),
        "append_only": ContainerSasPermissions(
            read=True, list=True, delete=False, write=True
        ),
        "full": ContainerSasPermissions(read=True, list=True, delete=True, write=True),
    }

    current_time = datetime.utcnow()

    signature = generate_container_sas(
        account_name,
        container_name,
        account_key=account_key,
        permission=access_policies[access_policy],
        start=current_time + timedelta(minutes=-1),
        expiry=current_time + expiry,
    )

    return container_url_template.format(
        account_name=account_name,
        container_name=container_name,
        signature=signature,
    )


class AzureCloudObjStorage(ObjStorage):
    """ObjStorage backend for Azure blob storage accounts.

    Args:
      container_url: the URL of the container in which the objects are stored.
      account_name: (deprecated) the name of the storage account under which objects are
        stored
      api_secret_key: (deprecated) the shared account key
      container_name: (deprecated) the name of the container under which objects are
        stored
      compression: the compression algorithm used to compress objects in storage
      use_secondary_endpoint_for_downloads: if True, use the secondary endpoint
        url to generate download URLs. To configure the secondary endpoint, use
        the BlobSecondaryEndpoint entry of the connection string.

    Notes:
      The container url should contain the credentials via a "Shared Access
      Signature". The :func:`get_container_url` helper can be used to generate
      such a URL from the account's access keys. The ``account_name``,
      ``api_secret_key`` and ``container_name`` arguments are deprecated.

    """

    PRIMARY_HASH = "sha1"
    name: str = "azure"

    def __init__(
        self,
        *,
        container_url: Optional[str] = None,
        account_name: Optional[str] = None,
        api_secret_key: Optional[str] = None,
        container_name: Optional[str] = None,
        connection_string: Optional[str] = None,
        compression: CompressionFormat = "gzip",
        use_secondary_endpoint_for_downloads=False,
        **kwargs,
    ):
        if container_url is None and connection_string is None:
            if account_name is None or api_secret_key is None or container_name is None:
                raise ValueError(
                    "AzureCloudObjStorage must have a container_url, a connection_string,"
                    "or all three account_name, api_secret_key and container_name"
                )
            else:
                warnings.warn(
                    "The Azure objstorage account secret key parameters are "
                    "deprecated, please use container URLs instead.",
                    DeprecationWarning,
                )
                container_url = get_container_url(
                    account_name=account_name,
                    account_key=api_secret_key,
                    container_name=container_name,
                    access_policy="full",
                )
        elif connection_string:
            if container_name is None:
                raise ValueError(
                    "container_name is required when using connection_string."
                )
            self.container_name = container_name

        super().__init__(**kwargs)
        self.container_url = container_url
        self.connection_string = connection_string
        self.compression = compression
        self.use_secondary = use_secondary_endpoint_for_downloads

    def get_container_client(self, hex_obj_id):
        """Get the container client for the container that contains the object with
        internal id hex_obj_id

        This is used to allow the PrefixedAzureCloudObjStorage to dispatch the
        client according to the prefix of the object id.

        """
        if self.connection_string:
            return ContainerClient.from_connection_string(
                self.connection_string, self.container_name
            )
        else:
            return ContainerClient.from_container_url(self.container_url)

    @contextlib.asynccontextmanager
    async def get_async_container_clients(self):
        """Returns a collection of container clients, to be passed to
        ``get_async_blob_client``.

        Each container may not be used in more than one asyncio loop."""
        if self.connection_string:
            client = AsyncContainerClient.from_connection_string(
                self.connection_string, self.container_name
            )
        else:
            client = AsyncContainerClient.from_container_url(self.container_url)
        async with client:
            yield {"": client}

    def get_blob_client(self, hex_obj_id):
        """Get the azure blob client for the given hex obj id"""
        container_client = self.get_container_client(hex_obj_id)

        return container_client.get_blob_client(blob=hex_obj_id)

    def get_async_blob_client(self, hex_obj_id, container_clients):
        """Get the azure blob client for the given hex obj id and a collection
        yielded by ``get_async_container_clients``."""

        return container_clients[""].get_blob_client(blob=hex_obj_id)

    def get_all_container_clients(self):
        """Get all active block_blob_services"""
        yield self.get_container_client("")

    def _internal_id(self, obj_id: ObjId) -> str:
        """Internal id is the hex version in objstorage."""
        if isinstance(obj_id, dict):
            obj_id = obj_id[self.PRIMARY_HASH]
        return hashutil.hash_to_hex(obj_id)

    def check_config(self, *, check_write):
        """Check the configuration for this object storage"""
        for container_client in self.get_all_container_clients():
            props = container_client.get_container_properties()

            # FIXME: check_write is ignored here
            if not props:
                return False

        return True

    @timed
    def __contains__(self, obj_id: ObjId) -> bool:
        """Does the storage contains the obj_id."""
        hex_obj_id = self._internal_id(obj_id)
        client = self.get_blob_client(hex_obj_id)
        try:
            client.get_blob_properties()
        except ResourceNotFoundError:
            return False
        else:
            return True

    def __iter__(self) -> Iterator[CompositeObjId]:
        """Iterate over the objects present in the storage."""
        for client in self.get_all_container_clients():
            for obj in client.list_blobs():
                yield {self.PRIMARY_HASH: hashutil.hash_to_bytes(obj.name)}

    def __len__(self):
        """Compute the number of objects in the current object storage.

        Returns:
            number of objects contained in the storage.

        """
        return sum(1 for i in self)

    @timed
    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        """Add an obj in storage if it's not there already."""
        if check_presence and obj_id in self:
            return

        hex_obj_id = self._internal_id(obj_id)

        # Send the compressed content
        data = self.compress(content)

        client = self.get_blob_client(hex_obj_id)
        try:
            client.upload_blob(data=data, length=len(data))
        except ResourceExistsError:
            # There's a race condition between check_presence and upload_blob,
            # that we can't get rid of as the azure api doesn't allow atomic
            # replaces or renaming a blob. As the restore operation explicitly
            # removes the blob, it should be safe to just ignore the error.
            pass

    def restore(self, content: bytes, obj_id: ObjId) -> None:
        """Restore a content."""
        if obj_id in self:
            self.delete(obj_id)

        return self.add(content, obj_id, check_presence=False)

    @timed
    def get(self, obj_id: ObjId) -> bytes:
        """retrieve blob's content if found."""
        return call_async(self._get_async, obj_id)

    async def _get_async(self, obj_id, container_clients=None):
        """Coroutine implementing ``get(obj_id)`` using azure-storage-blob's
        asynchronous implementation.
        While ``get(obj_id)`` does not need asynchronicity, this is useful to
        ``get_batch(obj_ids)``, as it can run multiple ``_get_async`` tasks
        concurrently."""
        if container_clients is None:
            # If the container_clients argument is not passed, create a new
            # collection of container_clients and restart the function with it.
            async with self.get_async_container_clients() as container_clients:
                return await self._get_async(obj_id, container_clients)

        hex_obj_id = self._internal_id(obj_id)
        client = self.get_async_blob_client(hex_obj_id, container_clients)

        try:
            download = await client.download_blob()
        except ResourceNotFoundError:
            raise ObjNotFoundError(obj_id) from None
        else:
            data = await download.content_as_bytes()

        return self.decompress(data, hex_obj_id)

    async def _get_async_or_none(self, obj_id, container_clients):
        """Like ``get_async(obj_id)``, but returns None instead of raising
        ResourceNotFoundError. Used by ``get_batch`` so other blobs can be returned
        even if one is missing."""
        try:
            return await self._get_async(obj_id, container_clients)
        except ObjNotFoundError:
            return None

    async def _get_batch_async(self, obj_ids):
        async with self.get_async_container_clients() as container_clients:
            return await asyncio.gather(
                *[
                    self._get_async_or_none(obj_id, container_clients)
                    for obj_id in obj_ids
                ]
            )

    def get_batch(self, obj_ids: Iterable[ObjId]) -> Iterator[Optional[bytes]]:
        """Retrieve objects' raw content in bulk from storage, concurrently."""
        return call_async(self._get_batch_async, obj_ids)

    def delete(self, obj_id: ObjId):
        """Delete an object."""
        super().delete(obj_id)  # Check delete permission
        hex_obj_id = self._internal_id(obj_id)
        client = self.get_blob_client(hex_obj_id)
        try:
            client.delete_blob()
        except ResourceNotFoundError:
            raise ObjNotFoundError(obj_id) from None

        return True

    def download_url(
        self,
        obj_id: ObjId,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        hex_obj_id = self._internal_id(obj_id)
        client = self.get_blob_client(hex_obj_id)
        try:
            client.get_blob_properties()
        except ResourceNotFoundError:
            raise ObjNotFoundError(obj_id)
        else:
            signature = generate_blob_sas(
                client.account_name,
                client.container_name,
                hex_obj_id,
                account_key=client.credential.account_key,
                permission=BlobSasPermissions(read=True),
                expiry=datetime.now() + (expiry or timedelta(hours=24)),
                content_disposition=content_disposition,
            )
            if self.use_secondary:
                return f"{client.secondary_endpoint}?{signature}"
            else:
                return f"{client.primary_endpoint}?{signature}"


class PrefixedAzureCloudObjStorage(AzureCloudObjStorage):
    """ObjStorage with azure capabilities, striped by prefix.

    accounts is a dict containing entries of the form:
        <prefix>: <container_url_for_prefix>
    """

    def __init__(
        self,
        accounts: Mapping[str, Union[str, Mapping[str, str]]],
        name: str = "azure-prefixed",
        compression: CompressionFormat = "gzip",
        **kwargs,
    ):
        # shortcut AzureCloudObjStorage __init__
        ObjStorage.__init__(
            self,
            name=name,
            **kwargs,
        )

        self.compression = compression

        # Definition sanity check
        prefix_lengths = set(len(prefix) for prefix in accounts)
        if not len(prefix_lengths) == 1:
            raise ValueError(
                "Inconsistent prefixes, found lengths %s"
                % ", ".join(str(lst) for lst in sorted(prefix_lengths))
            )

        self.prefix_len = prefix_lengths.pop()

        expected_prefixes = set(
            "".join(letters)
            for letters in product(
                set(string.hexdigits.lower()), repeat=self.prefix_len
            )
        )
        missing_prefixes = expected_prefixes - set(accounts)
        if missing_prefixes:
            raise ValueError(
                "Missing prefixes %s" % ", ".join(sorted(missing_prefixes))
            )

        do_warning = False

        self.container_urls = {}
        for prefix, container_url in accounts.items():
            if isinstance(container_url, dict):
                do_warning = True
                container_url = get_container_url(
                    account_name=container_url["account_name"],
                    account_key=container_url["api_secret_key"],
                    container_name=container_url["container_name"],
                    access_policy="full",
                )
            self.container_urls[prefix] = container_url

        if do_warning:
            warnings.warn(
                "The Azure objstorage account secret key parameters are "
                "deprecated, please use container URLs instead.",
                DeprecationWarning,
            )

    def get_container_client(self, hex_obj_id):
        """Get the block_blob_service and container that contains the object with
        internal id hex_obj_id
        """
        prefix = hex_obj_id[: self.prefix_len]
        return ContainerClient.from_container_url(self.container_urls[prefix])

    @contextlib.asynccontextmanager
    async def get_async_container_clients(self):
        # This is equivalent to:
        # client1 = AsyncContainerClient.from_container_url(url1)
        # ...
        # client16 = AsyncContainerClient.from_container_url(url16)
        # async with client1, ..., client16:
        #     yield {prefix1: client1, ..., prefix16: client16}
        clients = {
            prefix: AsyncContainerClient.from_container_url(url)
            for (prefix, url) in self.container_urls.items()
        }
        async with contextlib.AsyncExitStack() as stack:
            for client in clients.values():
                await stack.enter_async_context(client)
            yield clients

    def get_async_blob_client(self, hex_obj_id, container_clients):
        """Get the azure blob client for the given hex obj id and a collection
        yielded by ``get_async_container_clients``."""

        prefix = hex_obj_id[: self.prefix_len]
        return container_clients[prefix].get_blob_client(blob=hex_obj_id)

    def get_all_container_clients(self):
        """Get all active container clients"""
        # iterate on items() to sort blob services;
        # needed to be able to paginate in the list_content() method
        yield from (
            self.get_container_client(prefix) for prefix in sorted(self.container_urls)
        )



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
import logging
import sys
from time import monotonic
from typing import Tuple

from swh.core.api.gunicorn_config import *  # noqa

logger = logging.getLogger(__name__)


def worker_exit(arbiter, worker):
    logger.info("Calling worker_exit")
    shutdown_storage_backend()


def worker_int(worker):
    logger.warning("Calling worker_int")
    shutdown_storage_backend()


def shutdown_storage_backend():
    """Run on_shutdown callback for storage when a worker is terminating"""
    if "swh.objstorage.api.server" not in sys.modules:
        return
    objstorage = sys.modules["swh.objstorage.api.server"].objstorage

    if on_shutdown := getattr(objstorage, "on_shutdown", None):
        on_shutdown()


class ThrottledAccessLog(logging.Filter):
    """Throttle gunicorn access log lines for `status_codes` to at most one
    every `interval` seconds"""

    def __init__(self, interval: int = 60, status_codes: Tuple[int, ...] = (200,)):
        super().__init__(name="gunicorn.access")
        self.status_codes = set(str(code) for code in status_codes)
        self.endpoints: Counter[str] = Counter()
        self.interval = interval

        self.previous_flush = monotonic()
        self.deadline = self.previous_flush + self.interval

    def filter(self, record):
        # gunicorn.access records are using `s` for status code and `U` for the
        # requested path
        if not (ret := super().filter(record)):
            return ret

        if record.args["s"] not in self.status_codes:
            return True

        # If we quiesce different status codes, stick them in different buckets
        if len(self.status_codes) > 1:
            bucket = f"{record.args['U']}({record.args['s']})"
        else:
            bucket = record.args["U"]

        self.endpoints[bucket] += 1

        now = monotonic()
        if self.deadline > now:
            # Quiesce record
            return False

        # logging.Filter must mutate the log record instead of creating a
        # new one, up to and including Python 3.11
        record.msg = (
            "Served %(total)s requests in the last %(interval).1fs, "
            "including %(most_common)s"
        )
        record.args = {
            "total": self.endpoints.total(),
            "most_common": ", ".join(
                f"{v} {path}" for path, v in self.endpoints.most_common(3)
            ),
            "interval": now - self.previous_flush,
        }
        self.previous_flush = now
        self.deadline = now + self.interval
        self.endpoints = Counter()
        return True



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os

import pytest

from swh.model import hashutil
from swh.objstorage.constants import ID_DIGEST_LENGTH
from swh.objstorage.exc import ObjCorruptedError

from .objstorage_testing import ObjStorageTestFixture


class TestPathSlicingObjStorage(ObjStorageTestFixture):
    compression = "none"

    @pytest.fixture(autouse=True)
    def swh_objstorage_config(self, tmpdir):
        return {
            "cls": "pathslicing",
            "root": str(tmpdir),
            "slicing": "0:2/2:4/4:6",
            "compression": self.compression,
        }

    def content_path(self, obj_id):
        hex_obj_id = hashutil.hash_to_hex(obj_id)
        return self.storage.slicer.get_path(hex_obj_id)

    def test_iter(self):
        content, obj_id = self.hash_content(b"iter")
        assert not list(iter(self.storage))
        self.storage.add(content, obj_id=obj_id)
        assert list(iter(self.storage)) == [{self.storage.PRIMARY_HASH: obj_id}]

    def test_len(self):
        content, obj_id = self.hash_content(b"len")
        assert len(self.storage) == 0
        self.storage.add(content, obj_id=obj_id)
        assert len(self.storage) == 1

    def test_check_ok(self):
        content, obj_id = self.hash_content(b"check_ok")
        self.storage.add(content, obj_id=obj_id)
        assert self.storage.check(obj_id) is None

    def test_check_id_mismatch(self):
        _, obj_id = self.hash_content(b"check_id_mismatch")
        self.storage.add(b"unexpected content", obj_id=obj_id)
        with pytest.raises(ObjCorruptedError, match="Object corrupted"):
            self.storage.check(obj_id)

    def test_iterate_from(self):
        all_ids = []
        for i in range(100):
            content, obj_id = self.hash_content(b"content %d" % i)
            self.storage.add(content, obj_id=obj_id)
            all_ids.append({self.storage.PRIMARY_HASH: obj_id})
        all_ids.sort(key=lambda d: d[self.storage.PRIMARY_HASH])

        ids = list(self.storage.iter_from(b"\x00" * ID_DIGEST_LENGTH))
        assert ids == all_ids

        ids = list(self.storage.iter_from(all_ids[0]))
        assert ids == all_ids[1:]

        ids = list(self.storage.iter_from(all_ids[-1], n_leaf=True))
        n_leaf = ids[-1]
        ids = ids[:-1]
        assert n_leaf == 1
        assert len(ids) == 0

        ids = list(self.storage.iter_from(all_ids[-2], n_leaf=True))
        n_leaf = ids[-1]
        ids = ids[:-1]
        assert n_leaf == 2  # beware, this depends on the hash algo
        assert len(ids) == 1
        assert ids == all_ids[-1:]

    def test_fdatasync_default(self, mocker):
        content, obj_id = self.hash_content(b"check_fdatasync")
        patched = mocker.patch.multiple(
            "os", fsync=mocker.DEFAULT, fdatasync=mocker.DEFAULT
        )
        self.storage.add(content, obj_id=obj_id)
        if self.storage.use_fdatasync:
            assert patched["fdatasync"].call_count == 1
            assert patched["fsync"].call_count == 0
        else:
            assert patched["fdatasync"].call_count == 0
            assert patched["fsync"].call_count == 1

    def test_fdatasync_forced_on(self, mocker):
        self.storage.use_fdatasync = True
        content, obj_id = self.hash_content(b"check_fdatasync")
        patched = mocker.patch.multiple(
            "os", fsync=mocker.DEFAULT, fdatasync=mocker.DEFAULT
        )
        self.storage.add(content, obj_id=obj_id)
        assert patched["fdatasync"].call_count == 1
        assert patched["fsync"].call_count == 0

    def test_fdatasync_forced_off(self, mocker):
        self.storage.use_fdatasync = False
        content, obj_id = self.hash_content(b"check_fdatasync")
        patched = mocker.patch.multiple(
            "os", fsync=mocker.DEFAULT, fdatasync=mocker.DEFAULT
        )
        self.storage.add(content, obj_id=obj_id)
        assert patched["fdatasync"].call_count == 0
        assert patched["fsync"].call_count == 1

    def test_check_not_compressed_trailing_data(self):
        content, obj_id = self.hash_content(b"check_not_compressed")
        self.storage.add(content, obj_id=obj_id)
        with open(self.content_path(obj_id), "ab") as f:  # Add garbage.
            f.write(b"garbage")
        with pytest.raises(ObjCorruptedError, match="Object corrupted") as error:
            self.storage.check(obj_id)
        if self.compression != "none":
            assert "trailing data found" in error.value.args[0]

    def test_check_not_compressed(self):
        content, obj_id = self.hash_content(b"check_not_compressed")
        self.storage.add(content, obj_id=obj_id)
        with open(self.content_path(obj_id), "wb") as f:  # Replace by garbage.
            f.write(b"garbage")
        with pytest.raises(ObjCorruptedError, match="Object corrupted") as error:
            self.storage.check(obj_id)
        if self.compression != "none":
            assert "not a proper compressed file" in error.value.args[0]

    def test__iter__skip_tmpfile(self, caplog):
        content, obj_id = self.hash_content(b"skip_tmpfile")
        self.storage.add(content, obj_id=obj_id)
        path = self.storage.slicer.get_path(obj_id.hex())

        dir = os.path.dirname(path)
        bogus_path = os.path.join(dir, "bogus_file")
        with open(bogus_path, "w") as f:
            f.write("bogus")

        with caplog.at_level(logging.WARNING, "swh.objstorage.backends.pathslicing"):
            for _ in self.storage:
                pass

        assert len(caplog.records) == 1, [log.getMessage() for log in caplog.records]
        message = caplog.records[0].getMessage()
        assert "__iter__" in message
        assert bogus_path in message

    def test_iter_from_skip_tmpfile(self, caplog):
        content, obj_id = self.hash_content(b"skip_tmpfile")
        self.storage.add(content, obj_id=obj_id)
        path = self.storage.slicer.get_path(obj_id.hex())

        dir = os.path.dirname(path)
        bogus_path = os.path.join(dir, "bogus_file")
        with open(bogus_path, "w") as f:
            f.write("bogus")

        with caplog.at_level(logging.WARNING, "swh.objstorage.backends.pathslicing"):
            for _ in self.storage.iter_from(obj_id=b"\x00" * 20):
                pass

        assert len(caplog.records) == 1, [log.getMessage() for log in caplog.records]
        message = caplog.records[0].getMessage()
        assert "iter_from" in message
        assert bogus_path in message


class TestPathSlicingObjStorageGzip(TestPathSlicingObjStorage):
    compression = "gzip"


@pytest.mark.all_compression_methods
class TestPathSlicingObjStorageZlib(TestPathSlicingObjStorage):
    compression = "zlib"


@pytest.mark.all_compression_methods
class TestPathSlicingObjStorageBz2(TestPathSlicingObjStorage):
    compression = "bz2"


@pytest.mark.all_compression_methods
class TestPathSlicingObjStorageLzma(TestPathSlicingObjStorage):
    compression = "lzma"



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.objstorage.exc import (
    ObjCorruptedError,
    ObjNotFoundError,
    ReadOnlyObjStorageError,
)
from swh.objstorage.factory import get_objstorage
from swh.objstorage.objstorage import compute_hash


@pytest.fixture()
def objstorage(tmpdir):
    pstorage = {
        "cls": "pathslicing",
        "root": tmpdir,
        "slicing": "0:5",
    }
    base_storage = get_objstorage(**pstorage)
    storage = get_objstorage(
        "read-only",
        storage=base_storage,
    )
    return storage


@pytest.fixture()
def contentdata(objstorage):
    valid_content = b"pre-existing content"
    invalid_content = b"invalid_content"
    true_invalid_content = b"Anything that is not correct"
    absent_content = b"non-existent content"
    data = {
        "valid_content": valid_content,
        "valid_id": compute_hash(valid_content),
        "invalid_content": invalid_content,
        "true_invalid_content": true_invalid_content,
        "invalid_id": compute_hash(true_invalid_content),
        "absent_content": absent_content,
        "absent_id": compute_hash(absent_content),
    }
    # store one valid and one invalid content in the actual (backend) objstorage
    for cnt in ("valid", "invalid"):
        objstorage.storage.add(data[f"{cnt}_content"], obj_id=data[f"{cnt}_id"])
    return data


def test_can_contains(objstorage, contentdata):
    assert contentdata["valid_id"] in objstorage
    assert contentdata["invalid_id"] in objstorage
    assert contentdata["absent_id"] not in objstorage


def test_can_iter(objstorage, contentdata):
    assert {"sha1": contentdata["valid_id"]} in iter(objstorage)
    assert {"sha1": contentdata["invalid_id"]} in iter(objstorage)


def test_can_len(objstorage, contentdata):
    assert len(objstorage) == 2


def test_can_get(objstorage, contentdata):
    assert contentdata["valid_content"] == objstorage.get(contentdata["valid_id"])
    assert contentdata["invalid_content"] == objstorage.get(contentdata["invalid_id"])


def test_can_check(objstorage, contentdata):
    with pytest.raises(ObjNotFoundError):
        objstorage.check(contentdata["absent_id"])
    with pytest.raises(ObjCorruptedError):
        objstorage.check(contentdata["invalid_id"])
    objstorage.check(contentdata["valid_id"])


def test_cannot_add(objstorage, contentdata):
    with pytest.raises(ReadOnlyObjStorageError):
        objstorage.add(b"New content")
    with pytest.raises(ReadOnlyObjStorageError):
        objstorage.add(contentdata["valid_content"], contentdata["valid_id"])


def test_cannot_restore(objstorage, contentdata):
    with pytest.raises(ReadOnlyObjStorageError):
        objstorage.restore(contentdata["valid_content"], contentdata["valid_id"])


def test_check_config(objstorage):
    assert objstorage.check_config(check_write=False) is True
    assert objstorage.check_config(check_write=True) is False



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta
import io
from itertools import islice
import logging
from typing import Iterator, Optional
from urllib.parse import urlparse

from swh.model import hashutil
from swh.objstorage.backends.pathslicing import PathSlicer
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.interface import CompositeObjId, ObjId
from swh.objstorage.objstorage import (
    DEFAULT_LIMIT,
    CompressionFormat,
    ObjStorage,
    objid_to_default_hex,
    timed,
)

from .http import HttpFiler

LOGGER = logging.getLogger(__name__)


class SeaweedFilerObjStorage(ObjStorage):
    """ObjStorage with seaweedfs abilities, using the Filer API.

    https://github.com/chrislusf/seaweedfs/wiki/Filer-Server-API
    """

    PRIMARY_HASH = "sha1"
    name: str = "seaweedfs"

    def __init__(
        self,
        *,
        url: str = "",
        compression: CompressionFormat = "none",
        slicing: str = "",
        pool_maxsize: int = 100,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.compression = compression
        self.wf = HttpFiler(url, pool_maxsize=pool_maxsize)
        self.root_path = urlparse(url).path
        if not self.root_path.endswith("/"):
            self.root_path += "/"
        self.slicer = PathSlicer(self.root_path, slicing)

    def check_config(self, *, check_write):
        """Check the configuration for this object storage"""
        # FIXME: hopefully this blew up during instantiation
        return True

    @timed
    def __contains__(self, obj_id: ObjId) -> bool:
        return self.wf.exists(self._path(obj_id))

    def __iter__(self) -> Iterator[CompositeObjId]:
        """Iterate over the objects present in the storage

        Warning: Iteration over the contents of a cloud-based object storage
        may have bad efficiency: due to the very high amount of objects in it
        and the fact that it is remote, get all the contents of the current
        object storage may result in a lot of network requests.

        You almost certainly don't want to use this method in production.
        """
        for obj_id in self.list_content(limit=None):
            assert obj_id
            yield obj_id

    def __len__(self):
        """Compute the number of objects in the current object storage.

        Warning: this currently uses `__iter__`, its warning about bad
        performance applies.

        Returns:
            number of objects contained in the storage.

        """
        return sum(1 for i in self)

    @timed
    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        if check_presence and obj_id in self:
            return

        self.wf.put(io.BytesIO(self.compress(content)), self._path(obj_id))

    @timed
    def restore(self, content: bytes, obj_id: ObjId) -> None:
        return self.add(content, obj_id, check_presence=False)

    @timed
    def get(self, obj_id: ObjId) -> bytes:
        try:
            obj = self.wf.get(self._path(obj_id))
        except Exception as exc:
            LOGGER.info("Failed to get object %s: %r", self._path(obj_id), exc)
            raise ObjNotFoundError(obj_id)

        return self.decompress(obj, objid_to_default_hex(obj_id))

    def download_url(
        self,
        obj_id: ObjId,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        path = self._path(obj_id)
        if not self.wf.exists(path):
            raise ObjNotFoundError(obj_id)
        return self.wf.build_url(path)

    def delete(self, obj_id: ObjId):
        super().delete(obj_id)  # Check delete permission
        if obj_id not in self:
            raise ObjNotFoundError(obj_id)
        self.wf.delete(self._path(obj_id))
        return True

    def list_content(
        self,
        last_obj_id: Optional[ObjId] = None,
        limit: Optional[int] = DEFAULT_LIMIT,
    ) -> Iterator[CompositeObjId]:
        if last_obj_id:
            objpath = self._path(last_obj_id)
            startdir, lastfilename = objpath.rsplit("/", 1)
        else:
            startdir = self.root_path
            lastfilename = None
        for fname in islice(
            self.wf.iterfiles(startdir, last_file_name=lastfilename), limit
        ):
            bytehex = fname.rsplit("/", 1)[-1]
            yield {self.PRIMARY_HASH: hashutil.hash_to_bytes(bytehex)}

    # internal methods
    def _path(self, obj_id: ObjId):
        """Compute the backend path for the given obj id

        Given an object is, return the path part of the url to query the
        backend seaweedfs filer service with, according the configured path
        slicing.

        """
        return self.slicer.get_path(objid_to_default_hex(obj_id))



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.objstorage.factory import get_objstorage
from swh.objstorage.objstorage import ObjStorage


def test_instantiate_noop_objstorage():
    objstorage = get_objstorage(cls="noop")
    assert objstorage is not None
    assert isinstance(objstorage, ObjStorage)



# Copyright (C) 2016-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import closing
import secrets
import socket
from typing import Optional

from libcloud.common.types import InvalidCredsError
from libcloud.storage.providers import get_driver
from libcloud.storage.types import (
    ContainerDoesNotExistError,
    ObjectDoesNotExistError,
    Provider,
)
import pytest

from swh.objstorage.backends.libcloud import CloudObjStorage
from swh.objstorage.exc import ObjCorruptedError
from swh.objstorage.factory import OBJSTORAGE_IMPLEMENTATIONS
from swh.objstorage.objstorage import decompressors

from .objstorage_testing import ObjStorageTestFixture

API_KEY = "API_KEY"
API_SECRET_KEY = "API SECRET KEY"
CONTAINER_NAME = "test_container"


@pytest.fixture(scope="module", autouse=True)
def cloud_provider_mock():
    # register the cloud MockCloudObjStorage class defined in this file so the
    # helper test function get_cls() defined in objstorage_testing works OK and
    # ObjStorageTestFixture.test_name() is green here
    OBJSTORAGE_IMPLEMENTATIONS[
        "cloud"
    ] = f"{MockCloudObjStorage.__module__}.{MockCloudObjStorage.__name__}"
    try:
        yield
    finally:
        del OBJSTORAGE_IMPLEMENTATIONS["cloud"]


class MockLibcloudObject:
    """Libcloud object mock that replicates its API"""

    def __init__(self, name, content):
        self.name = name
        self.content = list(content)

    def as_stream(self):
        yield from iter(self.content)


class MockLibcloudDriver:
    """Mock driver that replicates the used LibCloud API"""

    def __init__(self, api_key, api_secret_key):
        self.containers = {CONTAINER_NAME: {}}  # Storage is initialized
        self.api_key = api_key
        self.api_secret_key = api_secret_key

    def _check_credentials(self):
        # Private method may be known as another name in Libcloud but is used
        # to replicate libcloud behavior (i.e. check credential at each
        # request)
        if self.api_key != API_KEY or self.api_secret_key != API_SECRET_KEY:
            raise InvalidCredsError()

    def get_container(self, container_name):
        try:
            return self.containers[container_name]
        except KeyError:
            raise ContainerDoesNotExistError(
                container_name=container_name, driver=self, value=None
            )

    def iterate_container_objects(self, container):
        self._check_credentials()
        yield from (v for k, v in sorted(container.items()))

    def get_object(self, container_name, obj_id):
        self._check_credentials()
        try:
            container = self.get_container(container_name)
            return container[obj_id]
        except KeyError:
            raise ObjectDoesNotExistError(object_name=obj_id, driver=self, value=None)

    def delete_object(self, obj):
        self._check_credentials()
        try:
            container = self.get_container(CONTAINER_NAME)
            container.pop(obj.name)
            return True
        except KeyError:
            raise ObjectDoesNotExistError(object_name=obj.name, driver=self, value=None)

    def upload_object_via_stream(self, content, container, obj_id):
        self._check_credentials()
        obj = MockLibcloudObject(obj_id, content)
        container[obj_id] = obj


class MockCloudObjStorage(CloudObjStorage):
    """Cloud object storage that uses a mocked driver"""

    def _get_driver(self, **kwargs):
        return MockLibcloudDriver(**kwargs)

    def _get_provider(self):
        # Implement this for the abc requirement, but behavior is defined in
        # _get_driver.
        pass


class TestCloudObjStorage(ObjStorageTestFixture):
    compression = "none"
    path_prefix: Optional[str] = None

    @pytest.fixture(autouse=True)
    def objstorage(self):
        self.storage = MockCloudObjStorage(
            CONTAINER_NAME,
            api_key=API_KEY,
            api_secret_key=API_SECRET_KEY,
            compression=self.compression,
            path_prefix=self.path_prefix,
        )

    def test_compression(self):
        content, obj_id = self.hash_content(b"add_get_w_id")
        self.storage.add(content, obj_id=obj_id)

        libcloud_object = self.storage._get_object(obj_id)
        raw_content = b"".join(libcloud_object.content)

        d = decompressors[self.compression]()
        assert d.decompress(raw_content) == content
        assert d.unused_data == b""

    def test_trailing_data_on_stored_blob(self):
        content, obj_id = self.hash_content(b"test content without garbage")
        self.storage.add(content, obj_id=obj_id)

        libcloud_object = self.storage._get_object(obj_id)
        libcloud_object.content.append(b"trailing garbage")

        with pytest.raises(ObjCorruptedError) as e:
            self.storage.check(obj_id)

        if self.compression != "none":
            assert "trailing data found" in e.value.args[0]

    @pytest.mark.skip("makes no sense to test this for the mocked libcloud")
    def test_download_url(self):
        pass


@pytest.mark.all_compression_methods
class TestCloudObjStorageBz2(TestCloudObjStorage):
    compression = "bz2"


class TestCloudObjStorageGzip(TestCloudObjStorage):
    compression = "gzip"


@pytest.mark.all_compression_methods
class TestCloudObjStorageLzma(TestCloudObjStorage):
    compression = "lzma"


@pytest.mark.all_compression_methods
class TestCloudObjStorageZlib(TestCloudObjStorage):
    compression = "zlib"


class TestCloudObjStoragePrefix(TestCloudObjStorage):
    path_prefix = "contents"

    def test_path_prefix(self):
        content, obj_id = self.hash_content(b"test content")
        self.storage.add(content, obj_id=obj_id)

        container = self.storage.driver.containers[CONTAINER_NAME]
        object_path = self.storage._object_path(obj_id)

        assert object_path.startswith(self.path_prefix + "/")
        assert object_path in container


def _find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


try:
    from moto.server import ThreadedMotoServer
except ImportError:
    moto_available = False
else:
    moto_available = True


@pytest.fixture(scope="class")
def moto_server():
    container_name = secrets.token_hex(10)
    port = _find_free_port()
    moto_config = {
        "key": "testing",
        "secret": "testing",
        "host": "localhost",
        "port": port,
        "secure": False,
    }
    server = ThreadedMotoServer(port=port)
    server.start()

    driver_cls = get_driver(Provider.S3)
    driver = driver_cls(**moto_config)
    container = driver.create_container(container_name=container_name)
    yield moto_config, driver, container
    driver.delete_container(container)
    server.stop()


@pytest.mark.skipif(not moto_available, reason="moto package is not installed")
class TestMotoS3CloudObjStorage(ObjStorageTestFixture):
    compression = "none"

    @pytest.fixture
    def swh_objstorage_config(self, moto_server):
        moto_config, driver, container = moto_server
        yield {
            "cls": "s3",
            "container_name": container.name,
            "compression": self.compression,
            **moto_config,
        }
        for obj in driver.list_container_objects(container):
            driver.delete_object(obj)
